// MPI distributed GA

#include <iostream>
#include "baseenv.h"
#include "mpicomm.h"
#include "benchtime.h"
#include "util.h"
#include "gafuncs.h"

using namespace std;

main(int argc, char *argv[])
{
	GA_GLOBAL_SETTINGS ggs;
	CyclicMigration cycricMig;
	MPICommunicator mc(cycricMig, &argc, &argv, "sample.ini");

	if(mc.wantHelp())
	{
		if(mc.getRank() == 0) cout << gappa_help;
		return 0;
	}

	const GA_GLOBAL_SETTINGS *pggs = mc.getGASettings();
	double_fitness_func func = rastrigin;
	double openMax = 5.12, closedMin = -5.12;
	if(strcmp(pggs->function_2_load, "schwefel") == 0)
		func = schwefel, openMax = 512, closedMin = -512;
	if(strcmp(pggs->function_2_load, "griewank") == 0)
		func = griewank, openMax = 512, closedMin = -512;
	if(strcmp(pggs->function_2_load, "rosenbrock") == 0)
		func = rosenbrock, openMax = 2.048, closedMin = -2.048;
	if(strcmp(pggs->function_2_load, "ridge") == 0)
		func = ridge, openMax = 64, closedMin = -64;

	BaseGAEnvironment env(func, &mc);

	// reset means no more environments to be create.
	mc.reset();

	env.setVarRange(closedMin, openMax);

	if(clParamExists(argc, argv, "sync"))
		mc.synchronizing(true);

	pggs = env.getGASettings();

	if(mc.getRank() == 0)
	{
		mpi_trace(
			"number of variables : %d\n"
			"design variable     : %d [%lf %lf)\n"
			"chromo length       : %d\n"
			"population size     : %d\n"
			"number of elite     : %d\n"
			"mutation rate       : %lf\n"
			"crossover rate      : %lf\n"
			"emigration rate     : %lf\n"
			"migration interval  : %d\n"
			"num of environments : %d\n"
			"process generations : %d\n"
			"number of trials    : %d\n",
			pggs->nVariables,
			pggs->dvd.nBit, pggs->dvd.closedMin, pggs->dvd.openMax,
			env.getChromoLengthByBit(),
			env.getPopulationSize(),
			env.getNumEliteSurvive(),
			env.getMutationRate(),
			env.getCrossoverRate(),
			env.getEmigrationRate(),
			env.getMigrationInterval(),
			env.getNumEnvironments(),
			pggs->nGenerations,
			pggs->nTrials);
	}

	bool fShow = clParamExists(argc, argv, "show");
	const char *chr[] = {"X", "|", "/", "-"};

	HighResTimer timer;
	int i, rank, trials;
	int tm = 0, anim = 0;
	for(trials = 1; trials <= pggs->nTrials; trials++)
	{
		if(rank) mpi_trace("\n\x1b[1A");
		env.establishEnvironment();
		mc.synchronize();
		mpi_trace("\x1b[s");
		rank = env.getEnvironmentId();
		mc.synchronize();

		timer.startCounter();

		double d = 0;
		bool f = false;
		for(i = 0; i < pggs->nGenerations; i++)
		{
			d = env.nextGeneration();
			if(abs(d) <= DBL_EPSILON)
			{
				mpi_trace("\x1b[u%s %2d,%5d : %+6.4lf\x1b[0K",
					chr[anim], trials, i, env(0).fitness());
				anim = (anim + 1) & 3;
				env.shutdown();
				f = true;
				break;
			}
			if(d == NG_SHUTDOWN_SIGNAL)
			{
				env.shutdown();
				break;
			}
			if(i % 50 == 0 && rank == 0)
			{
				mpi_trace("\x1b[u%s %2d,%5d : %+6.4lf\x1b[0K",
					chr[anim], trials, i, env(0).fitness());
				anim = (anim + 1) & 3;
			}
		}

		mc.synchronize();
		tm += timer.getCounter();

		if(f || (i == pggs->nGenerations && rank == 0))
		{
			mpi_trace("\x1b[uO %2d,%5d : %+6.4lf (%d)\n",
				trials, i, env(0).fitness(), tm / trials);
			cerr << env(0) << endl;
		}
		mc.synchronize();
	}

	return 0;
}
