//--------------------------------------------------------------------------
// baseenv.cpp
// GAPPA Parallel Distributed Environment Genetic Algorithm Library
// GA Common Task Layer
//--------------------------------------------------------------------------
// Copyright (C) 2000 Takashi Kawasaki, Tomoyuki Hiroyasu and Mitsunori Miki,
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright  notice
//    and this list of conditions.
// 2. Redistributions in binary form must reproduce the above  copyright notice
//    and this list of conditions in the documentation and/or other materials
//    provided with the distribution.
// 3. All advertising materials mentioning features or use of this software
//    must display the following acknowledgement:
//
//    This product includes software developed by Takashi Kawasaki,
//    Tomoyuki Hiroyasu and Mitsunori Miki except the functions of sgenrand()
//    and genrand().
//
// 4. The names of Takashi Kawasaki, Tomoyuki Hiroyasu and Mitsunori Miki
//    should not be used to endorse or promote products derived from this
//    software without specific prior written permission.

#include <iostream>
#include "baseenv.h"

using namespace std;

static const char *pformat_int = "%d";
static const char *pformat_double = "%6.4lf";
#if defined(_MSC_VER) // Visual C++
static const char *pformat_i64 = "%I64d";
#else // gcc/egcs, Borland C++
static const char *pformat_i64 = "%Ld";
#endif

//
// explicit constructors which associate fitness function and its type
//

BaseGAEnvironment::BaseGAEnvironment(const BGAE_func_adapter func_adapter, Communicator *com) : GeneEnvironment()
{
	memset(&ggs, 0, sizeof(GA_GLOBAL_SETTINGS));
	ggs.dvd.nBit = 10;
	ggs.dvd.closedMin = -5.12;
	ggs.dvd.closedMin = 5.12;
	ggs.nVariables = 10;
	ggs.nGenerations = 1000;
	ggs.nEnvironments = 1;
	ggs.function_2_load[0] = 0;
	ggs.fUseGrayCode = true;
	ggs.nTrials = 1;

	funcType = func_adapter.getFunctionType();
	func = func_adapter.getFunctionPtr();

	switch(funcType)
	{
	case ft_int:
		disp_format = pformat_int;
		break;

	case ft_double:
		disp_format = pformat_double;
		break;

	case ft_i64:
		disp_format = pformat_i64;
		break;
	}

	// set communicator
	comm = com;
	if(comm)
	{
		comm->associate(this);
		comm->inheritGASettings(&ggs);
	}
}

BaseGAEnvironment::~BaseGAEnvironment()
{
	if(comm)
		comm->disassociate(this);
}

//
// various type fitness functions
//
double BaseGAEnvironment::f_gene(const Gene& gen) const
{
	return ((gene_fitness_func)func)(gen);
}

double BaseGAEnvironment::f_bin(const Gene& gen) const
{
	return ((bin_fitness_func)func)
		(BitAccessor(const_cast<u8 *>(gen.getChromoPtr()),
		getChromoLengthByBit()));
}

double BaseGAEnvironment::f_int(const Gene& gen) const
{
	int *x = new int[ggs.nVariables];
	gen.extractDesignVariables(x, &ggs.dvd.nBit, ggs.nVariables,
		true, ggs.fUseGrayCode);
	double f = ((int_fitness_func)func)(x, ggs.nVariables);
	delete[] x;
	return f;
}

double BaseGAEnvironment::f_double(const Gene& gen) const
{
	double *x = new double[ggs.nVariables];
	gen.extractDesignVariables(x, &ggs.dvd, ggs.nVariables,
		true, ggs.fUseGrayCode);
	double f = ((double_fitness_func)func)(x, ggs.nVariables);
	delete[] x;
	return f;
}

#if defined(HAVE_64BIT_TYPE)
double BaseGAEnvironment::f_i64(const Gene& gen) const
{
	i64 *x = new i64[ggs.nVariables];
	gen.extractDesignVariables(x, &ggs.dvd.nBit, ggs.nVariables,
		true, ggs.fUseGrayCode);
	double f = ((i64_fitness_func)func)(x, ggs.nVariables);
	delete[] x;
	return f;
}
#endif

//
// fitness stub function
//
double BaseGAEnvironment::getFitness(const Gene& gen) const
{
	// most compiler can deal well with these type switch syntax,
	// they may replace this switch into simple dispatch table.
	switch(funcType)
	{
	case ft_gene:   return f_gene(gen);
	case ft_bin:    return f_bin(gen);
	case ft_int:    return f_int(gen);
	case ft_double: return f_double(gen);
#if defined(HAVE_64BIT_TYPE)
	case ft_i64:    return f_i64(gen);
#endif
	}
	return 0;
}

void BaseGAEnvironment::resetEnvironment()
{
	if(comm)
		comm->reset();

	GeneEnvironment::resetEnvironment();
	if(comm) comm->inheritGASettings(&ggs);
}

//
// configuration loader
//
bool BaseGAEnvironment::loadConfiguration(const char *filename, int n)
{
	if(!::loadConfiguration(&ggs, this, filename, n, true))
		return false;
	return setDisplayFormat(ggs.displayFormat);
}

void BaseGAEnvironment::setGASettings(const GA_GLOBAL_SETTINGS& rggs)
{
	if(&rggs == &ggs) return;
	ggs = rggs;
	setNumVariables(ggs.nVariables);
	setVarBitLength(ggs.dvd.nBit);
}

bool BaseGAEnvironment::setVarRange(double min, double max)
{
	if(max <= min)
		return false;

	ggs.dvd.openMax = max, ggs.dvd.closedMin = min;
	return true;
}

bool BaseGAEnvironment::setVarBitLength(int nBit)
{
	if(funcType == ft_gene || funcType == ft_bin)
		return false;

	if(funcType == ft_int && nBit >= sizeof(int) * 8)
		return false;

	if(funcType == ft_double && nBit > DBL_MANT_DIG)
		return false;

	if(funcType == ft_i64 && nBit >= 64)
		return false;

	ggs.dvd.nBit = nBit;
	setChromoLengthByBit(nBit * ggs.nVariables);
	return true;
}

bool BaseGAEnvironment::setNumVariables(int nVars)
{
	ggs.nVariables = nVars;
	setChromoLengthByBit(ggs.dvd.nBit * nVars);
	return true;
}

bool BaseGAEnvironment::loadCommandlineSetttings(int argc, const char * const argv[], bool fSetGlobalSettings)
{
	if(!::loadCommandlineSetttings(&ggs, this, argc, argv, fSetGlobalSettings))
		return false;
	setDisplayFormat(ggs.displayFormat);
	return true;
}

bool BaseGAEnvironment::loadSettings(int argc, const char * const argv[], const char *inifile, int n)
{
	if(!::loadSettings(&ggs, this, argc, argv, inifile, n, true))
		return false;
	setDisplayFormat(ggs.displayFormat);
	return true;
}

bool BaseGAEnvironment::setDisplayFormat(string format)
{
	if(check_disp_format(format.c_str()))
	{
		disp_format = format;
		return true;
	}
	return false;

}

//
// copy constructor, assign operator
//
BaseGAEnvironment::BaseGAEnvironment(const BaseGAEnvironment& env) : GeneEnvironment(env)
{
	memcpy(&ggs, &env.ggs, sizeof(GA_GLOBAL_SETTINGS));
	func = env.func;
	funcType = env.funcType;
}

BaseGAEnvironment& BaseGAEnvironment::operator=(const BaseGAEnvironment& env)
{
	this->GeneEnvironment::operator=(env);
	memcpy(&ggs, &env.ggs, sizeof(GA_GLOBAL_SETTINGS));
	func = env.func;
	funcType = env.funcType;
	return *this;
}

bool BaseGAEnvironment::run()
{
	int i, j;

	establishEnvironment();

	for(i = 0; i < ggs.nTrials; i++)
	{
		if(ggs.fCheckConvergence)
		{
			for(j = 0; j < ggs.nGenerations; j++)
				if(nextGeneration())
					break;
		}
		else
		{
			for(j = 0; j < ggs.nGenerations; j++)
				nextGeneration();
		}
	}
	return true;
}

string BaseGAEnvironment::get_string_bin(const Gene& gen) const
{
	string str = "";
	int i;

	for(i = 0; i < getChromoLengthByBit(); i++)
		if(gen(i)) str += "1"; else str += "0";
	return str;
}

string BaseGAEnvironment::get_string_int(const Gene& gen) const
{
	string str = "(";
	int i;
	char buf[64];

	int *x = new int[ggs.nVariables];
	gen.extractDesignVariables(x, &ggs.dvd.nBit, ggs.nVariables,
		true, ggs.fUseGrayCode);

	for(i = 0; i < ggs.nVariables - 1; i++)
	{
		sprintf(buf, disp_format.c_str(), x[i]);
		str = str + buf + ",";
	}

	sprintf(buf, disp_format.c_str(), x[i]);
	str = str + buf + ")";
	delete[] x;
	return str;
}

string BaseGAEnvironment::get_string_double(const Gene& gen) const
{
	string str = "(";
	int i;
	char buf[64];

	double *x = new double[ggs.nVariables];
	gen.extractDesignVariables(x, &ggs.dvd, ggs.nVariables,
		true, ggs.fUseGrayCode);

	for(i = 0; i < ggs.nVariables - 1; i++)
	{
		sprintf(buf, disp_format.c_str(), x[i]);
		str = str + buf + ",";
	}

	sprintf(buf, disp_format.c_str(), x[i]);
	str = str + buf + ")";
	delete[] x;
	return str;
}

#if defined(HAVE_64BIT_TYPE)
string BaseGAEnvironment::get_string_i64(const Gene& gen) const
{
	string str = "(";
	int i;
	char buf[64];

	i64 *x = new i64[ggs.nVariables];
	gen.extractDesignVariables(x, &ggs.dvd.nBit, ggs.nVariables,
		true, ggs.fUseGrayCode);

	for(i = 0; i < ggs.nVariables - 1; i++)
	{
		sprintf(buf, disp_format.c_str(), x[i]);
		str = str + buf + ",";
	}

	sprintf(buf, disp_format.c_str(), x[i]);
	str = str + buf + ")";
	delete[] x;
	return str;
}
#endif

string BaseGAEnvironment::get_string(const Gene& gen, GAType type) const
{
	if(type == ft_none) type = funcType;

	switch(type)
	{
	case ft_gene:
		return GeneEnvironment::getString(gen);

	case ft_bin:
		return get_string_bin(gen);

	case ft_int:
		return get_string_int(gen);

	case ft_double:
		return get_string_double(gen);

#if defined(HAVE_64BIT_TYPE)
	case ft_i64:
		return get_string_i64(gen);
#endif
	}
	return string("");
}

string BaseGAEnvironment::getString(const Gene& gen) const
{
	return get_string(gen, funcType);
}

bool BaseGAEnvironment::check_disp_format(const char *format) const
{
	// printf field format:
	// %[flags][width][.prec][F|N|h|l|L]type_char

	const char *p;

	// '%' exists ?
	p = strchr(format, '%');
	if(p == NULL)
		return false;

	// only one '%' ?
	if(strchr(p + 1, '%') != NULL)
		return false;

	p++;

	// flags
	if(*p == '-' || *p == '+' || *p == ' ' || *p == '#')
		p++;

	// width
	if(*p == '*')
		p++;
	else
		for(; *p; p++)
			if(*p < '0' || *p > '9') break;

	// prec
	if(*p == '.')
		for(p++; *p; p++)
			if(*p < '0' || *p > '9') break;

	// F|N|h|l|L
	if(strchr("FNhlL", *p)) p++;

	// type_char
	if(funcType == ft_double)
	{
		if(strchr("fegEG", *p)) // double
			return true;
	}
	else
	{
		if(strchr("diouxXc", *p)) // int
			return true;
	}
	return false;
}

bool BaseGAEnvironment::migrationProcedure(const constgene_pointer apEmigrants[], int nEmigrants)
{
	if(comm)
		return comm->migration(this, apEmigrants, nEmigrants);
	return false;
}

void BaseGAEnvironment::shutdown()
{
	if(comm)
		comm->shutdown();
}
