//--------------------------------------------------------------------------
// baseenv.h
// GAPPA Parallel Distributed Environment Genetic Algorithm Library
// GA Common Task Layer
//--------------------------------------------------------------------------
// Copyright (C) 2000 Takashi Kawasaki, Tomoyuki Hiroyasu and Mitsunori Miki,
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright  notice
//    and this list of conditions.
// 2. Redistributions in binary form must reproduce the above  copyright notice
//    and this list of conditions in the documentation and/or other materials
//    provided with the distribution.
// 3. All advertising materials mentioning features or use of this software
//    must display the following acknowledgement:
//
//    This product includes software developed by Takashi Kawasaki,
//    Tomoyuki Hiroyasu and Mitsunori Miki except the functions of sgenrand()
//    and genrand().
//
// 4. The names of Takashi Kawasaki, Tomoyuki Hiroyasu and Mitsunori Miki
//    should not be used to endorse or promote products derived from this
//    software without specific prior written permission.

#if !defined(_BASE_GA_ENVIRONMENT_)
#define _BASE_GA_ENVIRONMENT_ 1

#include <string>
#include "dega.h"
#include "util.h"
#include "communicator.h"

typedef double (*gene_fitness_func)(const Gene& gen);
typedef double (*bin_fitness_func)(const BitAccessor& bits);
typedef double (*int_fitness_func)(const int *x, int n);
typedef double (*double_fitness_func)(const double *x, int n);
#if defined(HAVE_64BIT_TYPE)
typedef double (*i64_fitness_func)(const i64 *x, int n);
#endif

typedef enum {ft_none, ft_gene, ft_bin, ft_int, ft_double, ft_i64} GAType;

// BGAE_func_adapter class checks function's prototype
class BGAE_func_adapter
{
	GAType type;
	void *func;
public:
	BGAE_func_adapter(gene_fitness_func f)   : func(f), type(ft_gene) {}
	BGAE_func_adapter(bin_fitness_func f)    : func(f), type(ft_bin) {}
	BGAE_func_adapter(int_fitness_func f)    : func(f), type(ft_int) {}
	BGAE_func_adapter(double_fitness_func f) : func(f), type(ft_double) {}
#if defined(HAVE_64BIT_TYPE)
	BGAE_func_adapter(i64_fitness_func f)    : func(f), type(ft_i64) {}
#endif
	inline void *getFunctionPtr() const {return func;}
	inline GAType getFunctionType() const {return type;}
};

// BaseGAEnvironment implementation
class BaseGAEnvironment : public GeneEnvironment
{
private:
	GA_GLOBAL_SETTINGS ggs;
	GAType funcType;
	void *func;
	std::string disp_format;
	Communicator *comm;

	double f_gene(const Gene& gen) const;
	double f_bin(const Gene& gen) const;
	double f_int(const Gene& gen) const;
	double f_double(const Gene& gen) const;
#if defined(HAVE_64BIT_TYPE)
	double f_i64(const Gene& gen) const;
#endif

	bool check_disp_format(const char *format) const;
	std::string get_string_bin(const Gene& gen) const;
	std::string get_string_int(const Gene& gen) const;
	std::string get_string_double(const Gene& gen) const;
#if defined(HAVE_64BIT_TYPE)
	std::string get_string_i64(const Gene& gen) const;
#endif
	std::string get_string(const Gene& gen, GAType type = ft_none) const;

	double getFitness(const Gene& gen) const;

public:
	explicit BaseGAEnvironment(const BGAE_func_adapter func_adapter, Communicator *com = NULL);
	BaseGAEnvironment(const BaseGAEnvironment& env);
	BaseGAEnvironment& operator=(const BaseGAEnvironment& env);
	virtual ~BaseGAEnvironment();
	virtual void resetEnvironment();

	bool loadConfiguration(const char *filename, int n = 0);
	bool loadCommandlineSetttings(int argc, const char * const argv[], bool fSetGlobalSettings = true);
	bool loadSettings(int argc, const char * const argv[], const char *inifile, int n = 0);
	bool setDisplayFormat(std::string format);
	bool run();
	const GA_GLOBAL_SETTINGS *getGASettings() const {return &ggs;}

	void setGASettings(const GA_GLOBAL_SETTINGS& rggs);
	double getVarOpenMax() const {return ggs.dvd.openMax;}
	double getVarClosedMin() const {return ggs.dvd.closedMin;}
	int getVarBitLength() const {return ggs.dvd.nBit;}
	int getNumVariables() const {return ggs.nVariables;}

	bool setVarRange(double min, double max);
	bool setVarBitLength(int nBit);
	bool setNumVariables(int nVars);

	virtual std::string getString(const Gene& gen) const;
	bool migrationProcedure(const constgene_pointer apEmigrants[], int nEmigrants);
	virtual void shutdown();
};

#endif // _BASE_GA_ENVIRONMENT_
