//--------------------------------------------------------------------------
// communicator.cpp
// GAPPA Parallel Distributed Environment Genetic Algorithm Library
// Distributed GA Communicator Object
//--------------------------------------------------------------------------
// Copyright (C) 2000 Takashi Kawasaki, Tomoyuki Hiroyasu and Mitsunori Miki,
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright  notice
//    and this list of conditions.
// 2. Redistributions in binary form must reproduce the above  copyright notice
//    and this list of conditions in the documentation and/or other materials
//    provided with the distribution.
// 3. All advertising materials mentioning features or use of this software
//    must display the following acknowledgement:
//
//    This product includes software developed by Takashi Kawasaki,
//    Tomoyuki Hiroyasu and Mitsunori Miki except the functions of sgenrand()
//    and genrand().
//
// 4. The names of Takashi Kawasaki, Tomoyuki Hiroyasu and Mitsunori Miki
//    should not be used to endorse or promote products derived from this
//    software without specific prior written permission.

#include <iostream>
#include <algorithm>
#include <cstring>
#include "communicator.h"
using namespace std;

void RandomDestMigration::resize(int n)
{
	if(sz == n) return;
	sz = n;
	if(dest) delete[] dest;
	dest = new int[sz];
	nextPattern();
}

void RandomDestMigration::nextPattern()
{
	int i;
	for(i = 0; i < sz; i++)
		dest[i] = genrand_int(sz);
}

Communicator::Communicator(MigrationRouter &mr, int argc, char **argv, const char *inifile) : router(&mr)
{
	if(argc && inifile)
		fShowHelp = !loadSettings(argc, argv, inifile);
	else
		fShowHelp = false;
}

SharedMemoryComm::SharedMemoryComm(MigrationRouter &mr, int _argc, char **_argv, const char *inifile) : Communicator(mr, _argc, _argv, inifile), gen(-1)
{
	argc = _argc;
	argv = _argv;
	if(inifile) strcpy(file, inifile); else file[0] = 0;
}

bool SharedMemoryComm::associate(GeneEnvironment *sender)
{
#if defined(_MT)
	mutex.lock();
#endif

	vector<GeneEnvironment *>::iterator it
		= find(envs.begin(), envs.end(), sender);
	if(it == envs.end() || it == NULL)
	{
		envs.push_back(sender);
		getRouter().resize(envs.size());
	}
#if defined(_MT)
	mutex.unlock();
#endif
	Communicator::associate(sender);
	return true;
}

bool SharedMemoryComm::disassociate(GeneEnvironment *sender)
{
#if defined(_MT)
	mutex.lock();
#endif

	GeneEnvironment * const * it
		= find(envs.begin(), envs.end(), sender);
	if(it != envs.end())
	{
		envs.erase(const_cast<GeneEnvironment **>(it));
		getRouter().resize(envs.size());
	}
#if defined(_MT)
	mutex.unlock();
#endif
	Communicator::disassociate(sender);
	return true;
}

void SharedMemoryComm::reset()
{
	int i;
	vector<GeneEnvironment *>::iterator it;
	for(it = envs.begin(); it < envs.end(); it++)
	{
		(*it)->setNumEnvironments(envs.size());
		(*it)->setEnvironmentId(*it - *envs.begin());
	}
}

int SharedMemoryComm::getEnvironmentId(GeneEnvironment *sender) const
{
#if defined(_MT)
	mutex.lock();
#endif
	GeneEnvironment * const * it
		= find(envs.begin(), envs.end(), sender);
	if(it == envs.end()) return -1;
#if defined(_MT)
	mutex.unlock();
#endif
	return (int)(it - envs.begin());
}

bool SharedMemoryComm::migration(GeneEnvironment *sender, const constgene_pointer apEmigrants[], int nEmigrants)
{
#if defined(_MT)
	mutex.lock();
#endif

	if(gen != sender->getTimer())
	{
		gen = sender->getTimer();
		getRouter().nextPattern();
	}

	GeneEnvironment *partner = envs[getRouter()[getEnvironmentId(sender)]];
#if defined(_MT)
	mutex.unlock();
#endif

#if defined(_MT)
	partner->lockVectorGeneInstance();
#endif

	int i;
	for(i = 0; i < nEmigrants; i++)
		partner->getVectorGeneInstance() = *apEmigrants[i];

#if defined(_MT)
	partner->releaseVectorGeneInstance();
#endif
	return false;
}
