//--------------------------------------------------------------------------
// communicator.h
// GAPPA Parallel Distributed Environment Genetic Algorithm Library
// Distributed GA Communicator Object
//--------------------------------------------------------------------------
// Copyright (C) 2000 Takashi Kawasaki, Tomoyuki Hiroyasu and Mitsunori Miki,
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright  notice
//    and this list of conditions.
// 2. Redistributions in binary form must reproduce the above  copyright notice
//    and this list of conditions in the documentation and/or other materials
//    provided with the distribution.
// 3. All advertising materials mentioning features or use of this software
//    must display the following acknowledgement:
//
//    This product includes software developed by Takashi Kawasaki,
//    Tomoyuki Hiroyasu and Mitsunori Miki except the functions of sgenrand()
//    and genrand().
//
// 4. The names of Takashi Kawasaki, Tomoyuki Hiroyasu and Mitsunori Miki
//    should not be used to endorse or promote products derived from this
//    software without specific prior written permission.
#ifndef _DGA_COMMUNICATOR_H_
#define _DGA_COMMUNICATOR_H_

#include <vector>

#if defined(_MT)
#include "muthread.h"
#endif

#include "dega.h"
#include "util.h"

class MigrationRouter
{
public:
	MigrationRouter() {}
	virtual ~MigrationRouter() {}
	virtual void resize(int n) = 0;
	virtual void srand(int n) = 0;
	virtual void nextPattern() = 0;
	virtual int operator[](int n) const = 0;
};

class CyclicMigration : public MigrationRouter
{
	ShuffledIntGenerator sig;
public:
	CyclicMigration() : MigrationRouter() {}
	void resize(int n) {sig.resize(n);}
	void srand(int n) {sig.srand(n);}
	void nextPattern() {sig.shuffle();}
	int operator[](int n) const {return sig[n];}
};

class RandomDestMigration : public MigrationRouter
{
	MTRand mt;
	int *dest;
	int sz;
public:
	RandomDestMigration() : MigrationRouter(), dest(NULL) {}
	void resize(int n);
	void srand(int n) {mt.srand(n);}
	void nextPattern();
	int operator[](int n) const {return dest[n];}
};

class ShiftMigration : public MigrationRouter
{
	int sz;
public:
	ShiftMigration() : MigrationRouter() {}
	void resize(int n) {sz = n;}
	void srand(int n) {}
	void nextPattern() {}
	int operator[](int n) const {return (n + 1) % sz;}
};

class Communicator
{
private:
	MigrationRouter *router;
	GA_GLOBAL_SETTINGS ggs;
	bool fShowHelp;

protected:
	inline MigrationRouter& getRouter() {return *router;}
public:
	explicit Communicator(MigrationRouter &mr, int argc = 0, char **argv = NULL, const char *inifile = NULL);
	bool loadSettings(int argc = 0, char **argv = NULL, const char *inifile = NULL) {return ::loadSettings(&ggs, NULL, argc, argv, inifile);}
	virtual ~Communicator() {}
	virtual void reset() = 0;
	virtual void shutdown() {}
	virtual int getNumEnvironments() const = 0;
	virtual bool associate(GeneEnvironment *sender) {return true;}
	virtual bool disassociate(GeneEnvironment *sender) {return true;}
	virtual int getEnvironmentId(GeneEnvironment *sender) const = 0;

	virtual bool migration(GeneEnvironment *sender, const constgene_pointer apEmigrants[], int nEmigrants) = 0;

	void inheritGASettings(GA_GLOBAL_SETTINGS *pggs) const
		{memcpy(pggs, &ggs, sizeof(GA_GLOBAL_SETTINGS));}
	bool wantHelp() const {return fShowHelp;}
	void wantHelp(bool f) {fShowHelp = f;}
	const GA_GLOBAL_SETTINGS *getGASettings() const {return &ggs;}
private:
	// copy of instance is not allowed
	Communicator(const Communicator& comm) {}
	Communicator& operator=(const Communicator& comm) {return *this;}
};

class SharedMemoryComm : public Communicator
{
private:
	int argc;
	char **argv;
	char file[256];
	std::vector<GeneEnvironment *> envs;
	mutable int gen;
#if defined(_MT)
	mutable muMutex mutex;
#endif

public:
	SharedMemoryComm(MigrationRouter &mr, int argc = 0, char **pargv = NULL, const char *inifile = NULL);
	~SharedMemoryComm() {}
	int getNumEnvironments() const {return envs.size();}
	bool associate(GeneEnvironment *sender);
	bool disassociate(GeneEnvironment *sender);
	int getEnvironmentId(GeneEnvironment *sender) const;
	void reset();
	bool migration(GeneEnvironment *sender, const constgene_pointer apEmigrants[], int nEmigrants);
};

#endif // _DGA_COMMUNICATOR_H_
