//--------------------------------------------------------------------------
// dega.cpp
// GAPPA Parallel Distributed Environment Genetic Algorithm Library
// DEGA main logic
//--------------------------------------------------------------------------
// Copyright (C) 2000 Takashi Kawasaki, Tomoyuki Hiroyasu and Mitsunori Miki,
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright  notice
//    and this list of conditions.
// 2. Redistributions in binary form must reproduce the above  copyright notice
//    and this list of conditions in the documentation and/or other materials
//    provided with the distribution.
// 3. All advertising materials mentioning features or use of this software
//    must display the following acknowledgement:
//
//    This product includes software developed by Takashi Kawasaki,
//    Tomoyuki Hiroyasu and Mitsunori Miki except the functions of sgenrand()
//    and genrand().
//
// 4. The names of Takashi Kawasaki, Tomoyuki Hiroyasu and Mitsunori Miki
//    should not be used to endorse or promote products derived from this
//    software without specific prior written permission.

#include <iostream>
#include <string>

#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <cmath>
#include "dega.h"

using namespace std;

void err_print(const char *format, ...)
{
	va_list argptr;
	char buffer[256];
	int cnt;

	va_start(argptr, format);
	vsprintf(buffer, format, argptr);
	va_end(argptr);

	strcat(buffer, "\n");
	cerr << buffer;
}


// binary search of a range which contains the value r
static int bin_search(double *ad, double r, int first, int end)
{
	if(end - first <= 1)
	{
		if(ad[first] <= r && r < ad[first + 1])
			return first;
	}

	int m = (first + end) / 2;
	if(r < ad[m])
		return bin_search(ad, r, first, m);
	else
		return bin_search(ad, r, m, end);
}

// default fitness scaling function
double defaultScalingFunc(double fitness, double max, double min)
{
	return fitness - min;
}

//////////////////////////////////////////////////////////////////////////////
// Gene
//

// internal constructor
void Gene::construct(const GeneEnvironment& environment)
{
	env = &environment;
	b = new u8[env->getChromoLength()];
	clearFitnessCache();

#if defined(USE_COLOR_PROFILING)
	colors = new u8[env->getSizeOfEnvironmentColoringStr()];
#endif
}

void Gene::destruct()
{
	if(b) {delete[] b; b = NULL;}
#if defined(USE_COLOR_PROFILING)
	if(colors) {delete[] colors; colors = NULL;}
#endif
}

// acquired re-constructor
// re-initialize the instance which has been already initialized.
// This method associate a new environment to a instance
// [CAUTION] This method broke his personality !
void Gene::acquireCitizenship(const GeneEnvironment& environment)
{
	destruct();
	construct(environment);
#if defined(USE_COLOR_PROFILING)
	setEnvironmentColor();
#endif
}

int Gene::getChromoLength() const
{
	return env->getChromoLength();
}

int Gene::getChromoLengthByBit() const
{
	return env->getChromoLengthByBit();
}

// naturalize is a method to change Gene's environment without restructuring
// his personality.
// this function failed if the length of chromosome is different between here
// (my environment) and there(that environment).
bool Gene::naturalize(const GeneEnvironment& environment)
{
	if(env->getChromoLength() != environment.getChromoLength())
		return false;
	env = &environment;
	return true;
}

// general constructor
Gene::Gene(GeneEnvironment& environment)
{
	construct(environment);
#if defined(USE_COLOR_PROFILING)
	setEnvironmentColor();
#endif
}

// default constructor, do nothing but confirm that it isn't initialized
Gene::Gene()
{
	env = NULL;
	b = NULL;
#if defined(USE_COLOR_PROFILING)
	colors = NULL;
#endif
	clearFitnessCache();
}

Gene::~Gene()
{
	destruct();
}

Gene::Gene(const Gene& gen)
{
	copy(gen);
}

#if defined(USE_COLOR_PROFILING)
void Gene::setEnvironmentColor()
{
	int i;
	for(i = 0; i < env->getSizeOfEnvironmentColoringStr(); i++)
	{
		if(env->getEnvironmentId() == i)
			colors[i] = 0xff;
		else
			colors[i] = 0;
	}
}
#endif

void Gene::copy(const Gene& gen)
{
	if(gen.b)
	{
		if(env == gen.env)
		{
			if(b == NULL)
				b = new u8[env->getChromoLength()];
		}
		else
		{
			if(b) delete[] b;
			b = new u8[gen.env->getChromoLength()];
		}
		memcpy(b, gen.b, gen.env->getChromoLength());
	}
	else if(b)
	{
		delete[] b;
		b = NULL;
	}

	env = gen.env;

#if defined(USE_COLOR_PROFILING)
	if(gen.colors)
	{
		if(colors) delete[] colors;
		colors = new u8[getSizeOfEnvironmentColoringStr()];
		memcpy(colors, gen.colors, getSizeOfEnvironmentColoringStr());
	}
	else if(colors)
	{
		delete[] colors;
		colors = NULL;
	}
#endif
#if !defined(NO_FITNESS_CACHE)
	dFitness = gen.dFitness;
#endif
}

// randomize
void Gene::randomize()
{
	int i;
	int gap = env->getChromoLengthByBit() & 7;
	for(i = 0; i < env->getChromoLengthByBit() / 8; i++)
		b[i] = (u8)genrand_int(256);
	if(gap)
		b[i] = (u8)genrand_int(1 << gap) << (8 - gap);
	clearFitnessCache();
}

// set bit
void Gene::setBit(int n, bool f)
{
	if(f) b[n >> 3] |= 128 >> (n & 7);
	else  b[n >> 3] &= (-1 ^ (128 >> (n & 7)));
	clearFitnessCache();
}

void Gene::mutate()
{
#if !defined(OPTIMIZATION_FOR_SPEED)
// Normal implementation
// just check whether it does mutate or not for each bit.
	int i;
#if !defined(NO_FITNESS_CACHE)
	bool f = false;
#endif
	for(i = 0; i < env->getChromoLengthByBit(); i++)
		if(env->genrand() < env->getMutationRate())
		{
			b[i >> 3] ^= 128 >> (i & 7);
		#if !defined(NO_FITNESS_CACHE)
			f = true;
		#endif
		}
#if !defined(NO_FITNESS_CACHE)
	if(f) clearFitnessCache();
#endif

#else // OPTIMIZATION_FOR_SPEED
// Faster version (16 times faster)
// this function estimates the number of mutation points by probability
// distribution function (binary search). and simply mutate these bits.
	int i, j, n;
	n = env->getNumMutationPoints();
	if(n)
	{
#if !defined(NO_FITNESS_CACHE)
		clearFitnessCache();
#endif
		for(i = 0; i < n; i++)
		{
			j = genrand_int(env->getChromoLengthByBit());
			b[j >> 3] ^= 128 >> (j & 7);
		}
	}
#endif // OPTIMIZATION_FOR_SPEED
}

// Exchange information about Gene each other
// this enables fast implementation of exchanging gene between env and env.
bool Gene::exchange(Gene& gen)
{
	u8 *temp = b;
	b = gen.b;
	gen.b = temp;

#if defined(USE_COLOR_PROFILING)
	temp = colors;
	colors = gen.colors;
	gen.colors = temp;
#endif

	const GeneEnvironment *tempEnv = env;
	env = gen.env;
	gen.env = tempEnv;

	return true;
}

void crossover(const u8 *parent1, const u8 *parent2, u8 *child1, u8 *child2, int size, int size_bit, int num_pts)
{
	u8 *bitmask = new u8[size];
	int i;
	if(num_pts == 0)
	{
		// ZERO POINT CROSS OVER as UNIFORM CROSS OVER
		for(i = 0; i < size; i++)
			bitmask[i] = (u8)genrand_int(256);
	}
	else if(num_pts == 1)
	{
		// create bitmask for 1point crossover
		int pt = genrand_int(size_bit);
		for(i = 0; i < pt / 8; i++)
			bitmask[i] = 0;
		bitmask[i++] = 0xff >> (pt & 7);
		for(; i < size; i++)
			bitmask[i] = 0xff;
	}
	else if(num_pts == 2)
	{
		// create bitmask for 2point crossover
		int pt[2];
		do{
			pt[0] = genrand_int(size_bit);
			pt[1] = genrand_int(size_bit);
		}
		while(pt[0] < pt[1]);

		for(i = 0; i < size; i++)
			bitmask[i] = 0xff;

		for(i = 0; i < num_pts; i++)
		{
			int n = pt[i];
			bitmask[n / 8] ^= 0xff >> (n & 7);
			for(n = n / 8 + 1; n < size; n++)
				bitmask[n] ^= 0xff;
		}
	}
	else
	{
		// create bitmask for n-point crossover
		int n;
		ShuffledIntGenerator sig(size_bit);

		for(i = 0; i < size; i++)
			bitmask[i] = 0xff;

		for(i = 0; i < num_pts; i++)
		{
			n = sig[i];
			bitmask[n / 8] ^= 0xff >> (n & 7);
			for(n = n / 8 + 1; n < size; n++)
				bitmask[n] ^= 0xff;
		}
	}

	// crossing over main
	for(i = 0; i < size; i++)
	{
		child1[i] =
			(u8)(parent1[i] &  bitmask[i]) | (u8)(parent2[i] & ~bitmask[i]);
		child2[i] =
			(u8)(parent1[i] & ~bitmask[i]) | (u8)(parent2[i] &  bitmask[i]);
	}
	delete[] bitmask;
}

void Gene::crossover(const Gene& partner, Gene& child1, Gene& child2) const
{
	::crossover(b, partner.b, child1.b, child2.b,
		env->getChromoLength(), env->getChromoLengthByBit(),
		env->getNumCrossoverPoints());

#if !defined(NO_FITNESS_CACHE)
	child1.clearFitnessCache();
	child2.clearFitnessCache();

#if defined(FORCE_CHROMO_STRICT_CHECK) // forch strict check
	if(memcmp(b, child1.b, env->getChromoLength()) == 0)
		child1.dFitness = dFitness;
	else if(memcmp(partner.b, child1.b, env->getChromoLength()) == 0)
		child1.dFitness = partner.dFitness;

	if(memcmp(b, child2.b, env->getChromoLength()) == 0)
		child2.dFitness = dFitness;
	else if(memcmp(partner.b, child2.b, env->getChromoLength()) == 0)
		child2.dFitness = partner.dFitness;
#endif
#endif

#if defined(USE_COLOR_PROFILING)
	// hybrid of WHITE and BLACK is GRAY
	for(i = 0; i < env->getSizeOfEnvironmentColoringStr(); i++)
	{
		child1.colors[i] = (u8)(((int)colors[i] + (int)partner.colors[i]) / 2);
		child2.colors[i] = child1.colors[i];
	}
#endif
}

// Gene::fitness is somewhat faster than GeneEnvironment::getFitness
// because of the cache.
// But if fitness function is easy to evaluate realtimely,
// you may use getFitness for the purpose of speed up
double Gene::fitness() const
{
#if !defined(NO_FITNESS_CACHE)
	if(dFitness != DBL_MAX)
		return dFitness;

	dFitness = env->getFitness(*this);
	return dFitness;
#else
	return env->getFitness(*this);
#endif
}

// gray coding
template<class T> T gray2num(T x)
{
	T mask;
	bool f = false;
	for(mask = T(1) << (sizeof(T) * 8 - 2); mask != 0; mask >>= 1)
	{
		if(f)
		{
			if(x & mask) f = false;
			x ^= mask;
		}
		else
			if(x & mask) f = true;
	}
	return x;
}

template<class T> T num2gray(T x)
{
	T mask;
	bool f = false;
	for(mask = T(1) << (sizeof(T) * 8 - 2); mask != 0; mask >>= 1)
	{
		if(f)
		{
			x ^= mask;
			if(x & mask) f = false;
		}
		else
			if(x & mask) f = true;
	}
	return x;
}

// subcontractor function of extractDesignVariables
template<class T> int read_bit_sequence(T *data, const u8 **pp, int bit_offset, int bit_2_read)
{
	const u8 *p = *pp;
	int w = 8 - bit_offset;

#if !defined(GENE_LONGER_THAN_8BIT)
	// Most people does'nt need this procedure, but complete implementation
	// must have this. IF YOU DON'T WANT TO DO THIS, SET DF_LONGER_THAN_8BIT
	// the number of bits to read is less than 8bit
	if(bit_2_read < w)
	{
		*data = (*p >> (w - bit_2_read)) & ((1 << bit_2_read) - 1);
		return bit_offset + bit_2_read;
	}
#endif

	// read ooooXXXX
	T n = *p++ & ((1 << w) - 1);
	bit_2_read -= w;

	// read XXXXXXXX several times
	for(;bit_2_read >= 8; bit_2_read -= 8)
		n = (n << 8) | *p++;

	// read XXoooooo
	if(bit_2_read)
	{
		bit_offset = 8 - bit_2_read;
		*data = (n << bit_2_read) | (*p >> bit_offset);
		*pp = p;
		return bit_2_read;
	}

	// data is simply 8bit aligned, no odd bits was read.
	*data = n;
	*pp = p;
	return 0;
}

// subcontractor function of buildDesignVariables
template<class T> int write_bit_sequence(T n, u8 **pp, int bit_offset, int nBits)
{
	u8 *p = *pp;

	int nb_start = 8 - bit_offset;
	int nb_end   = (nBits - nb_start) & 7;
	int nb_main  = (nBits - nb_start - nb_end);
	int nbyte_main  = nb_main / 8;

	// last part
	p[nbyte_main + 1] |= (n & ((1 << nb_end) - 1)) << (8 - nb_end);
	*pp = p + nbyte_main + 1;
	n >>= nb_end;

	// main part
	int i;
	for(i = nbyte_main; i > 0; i--)
	{
		p[i] = n & 255;
		n >>= 8;
	}

	// first part
	p[0] |= n & ((1 << nb_start) - 1);

	return nb_end;
}

// these method extracts design variables (as int value) from chromosome.
// anDesignVariable : the array of int on which return value will be written
// anBitSize        : the array of int which specifies the size of each design
//                    values.
// if fBalanced is false:
//   anBitSize array must have at least the size specified by nDesignVariable
//   each design variable has its own size.
// else (true):
//   anBitSize must be a pointer to a int variable which contains bit size
//   of each design variable : all design variable have same size.
// fUseGrayCode     : whether you want to use gray coding or not.
//                    default is true.
bool Gene::extractDesignVariables(int anDesignVariable[], const int *anBitSize, int nDesignVariable, bool fBalanced, bool fUseGrayCode) const
{
	int i, n, m, bit_offset;
	const u8 *p = b;
	for(i = 0, m = 0, bit_offset = 0; i < nDesignVariable; i++)
	{
		bit_offset = read_bit_sequence(&n, &p, bit_offset, anBitSize[m]);
		if(fUseGrayCode)
			anDesignVariable[i] = gray2num(n);
		else
			anDesignVariable[i] = n;

		// if not balanced
		if(!fBalanced) m++;
	}
	return true;
}

#if defined(HAVE_64BIT_TYPE)
bool Gene::extractDesignVariables(i64 anDesignVariable[], const int *anBitSize, int nDesignVariable, bool fBalanced, bool fUseGrayCode) const
{
	int i, m, bit_offset;
	i64 n;
	const u8 *p = b;
	for(i = 0, m = 0, bit_offset = 0; i < nDesignVariable; i++)
	{
		bit_offset = read_bit_sequence(&n, &p, bit_offset, anBitSize[m]);
		if(fUseGrayCode)
			anDesignVariable[i] = gray2num(n);
		else
			anDesignVariable[i] = n;

		// if not balanced
		if(!fBalanced) m++;
	}
	return true;
}
#endif // HAVE_64BIT_TYPE

bool Gene::extractDesignVariables(double adDesignVariable[], const DesignVariableDomain *dvd, int nDesignVariable, bool fBalanced, bool fUseGrayCode) const
{
	int i, m, bit_offset;
#if defined(OPTIMIZED_FOR_PRECISION) && defined(HAVE_64BIT_TYPE)
		i64 n;
#else
		int n;
#endif
	const u8 *p = b;
	for(i = 0, m = 0, bit_offset = 0; i < nDesignVariable; i++)
	{
		bit_offset = read_bit_sequence(&n, &p, bit_offset, dvd[m].nBit);
		if(fUseGrayCode) n = gray2num(n);
		adDesignVariable[i] = (double)n * (dvd[m].openMax - dvd[m].closedMin)
			/ (double)(1 << dvd[m].nBit) + dvd[m].closedMin;

		// if not balanced
		if(!fBalanced) m++;
	}
	return true;
}

// these method constructs design variables (as int value) on a chromosome.
// anDesignVariable : the array of int by which the chromosome is built
// anBitSize        : the array of int which specifies the size of each design
//                    values.
// if fBalanced is false:
//   anBitSize array must have at least the size specified by nDesignVariable
//   each design variable has its own size.
// else (true):
//   anBitSize must be a pointer to a int variable which contains bit size
//   of each design variable : all design variable have same size.
// fUseGrayCode     : whether you want to use gray coding or not.
//                    default is true.
bool Gene::buildDesignVariables(const int anDesignVariable[], const int *anBitSize, int nDesignVariable, bool fBalanced, bool fUseGrayCode)
{
	u8 *p = b;
	int i, bit_offset = 0;
	for(i = 0; i < nDesignVariable; i++)
	{
		int n;
		if(fUseGrayCode)
			n = num2gray(anDesignVariable[i]);
		else
			n = anDesignVariable[i];
		bit_offset = write_bit_sequence(n, &p, bit_offset, *anBitSize);
		if(!fBalanced) anBitSize++;
	}
	dFitness = DBL_MAX;
	return true;
}

#if defined(OPTIMIZED_FOR_PRECISION) && defined(HAVE_64BIT_TYPE)
bool Gene::buildDesignVariables(const i64 anDesignVariable[], const int *anBitSize, int nDesignVariable, bool fBalanced, bool fUseGrayCode)
{
	u8 *p = b;
	int i, bit_offset = 0;
	for(i = 0; i < nDesignVariable; i++)
	{
		i64 n;
		if(fUseGrayCode)
			n = num2gray(anDesignVariable[i]);
		else
			n = anDesignVariable[i];
		bit_offset = write_bit_sequence(n, &p, bit_offset, *anBitSize);
		if(!fBalanced) anBitSize++;
	}
	dFitness = DBL_MAX;
	return true;
}
#endif

bool Gene::buildDesignVariables(const double adDesignVariable[], const DesignVariableDomain *dvd, int nDesignVariable, bool fBalanced, bool fUseGrayCode)
{
	u8 *p = b;
	int i, bit_offset = 0;
	for(i = 0; i < nDesignVariable; i++)
	{
#if defined(OPTIMIZED_FOR_PRECISION) && defined(HAVE_64BIT_TYPE)
		i64 n;
#else
		int n;
#endif
		n = (int)((adDesignVariable[i] - dvd->closedMin)
			* (1 << dvd->nBit) / (dvd->openMax - dvd->closedMin));
		if(fUseGrayCode) n = num2gray(n);
		bit_offset = write_bit_sequence(n, &p, bit_offset, dvd->nBit);
		if(!fBalanced) dvd++;
	}
	dFitness = DBL_MAX;
	return true;
}

static const char ch[] = "0123456789ABCDEF";
// This function converts binary chromosome data into visible HEX(16) string
// buffer : the pointer to the buffer
// size   : specifies the length of buffer in bytes
const char* Gene::extractHexStr(char *buffer, int size) const
{
	char *p = buffer;
	if(size < getChromoLength() * 2 + 1)
		return NULL;

	int i;
	for(i = 0; i < getChromoLength(); i++)
	{
		*p++ = ch[b[i] / 16];
		*p++ = ch[b[i] & 15];
	}
	*p = 0;
	return buffer;
}

#if defined(USE_COLOR_PROFILING)
const char* Gene::extractColoringStr(char *buffer, int size) const
{
	char *p = buffer;
	if(size < env->getSizeOfEnvironmentColoringStr() * 2)
		return NULL;

	int i;
	for(i = 0; i < env->getSizeOfEnvironmentColoringStr(); i++)
	{
		*p++ = ch[colors[i] / 16];
		*p++ = ch[colors[i] & 15];
		*p++ = ',';
	}
	p--;
	*p = 0;
	return buffer;
}

int Gene::getColorLength() const
{
	return env->getSizeOfEnvironmentColoringStr();
}
#endif

int Gene::getRnaLength() const
{
	return env->getRnaLength();
}

const u8* Gene::transcription(u8 *buffer) const
{
	if(env == NULL) return NULL;

	int n;

#if defined(SEND_FITNESS_OVER_NETWORK)
	if(dFitness != DBL_MAX)
	{
		i32 mant = (i32)(frexp(dFitness, &n) * (1 << 30));

		u32 mant_n = htonl(mant);
		u32 exp2   = htonl(n);

		memcpy(buffer, &mant_n, 4);
		memcpy(buffer + 4, &exp2, 4);
	}
	else
		memset(buffer, 0xff, 8);
#endif

	n = env->getChromoLength();
	buffer[RNA_OFFSET] = n & 255;
	buffer[RNA_OFFSET + 1] = n / 256;
	memcpy(buffer + RNA_OFFSET + 2, b, n);

#if defined(USE_COLOR_PROFILING)
	n += 2;
	buffer[RNA_OFFSET + n] = env->getSizeOfEnvironmentColoringStr() & 255;
	buffer[RNA_OFFSET + n + 1] = env->getSizeOfEnvironmentColoringStr() / 256;
	memcpy(buffer + n + RNA_OFFSET + 2, colors,
		env->getSizeOfEnvironmentColoringStr());
#endif
	return buffer;
}

const u8 *extractRnaInformation(const u8 *data, int *pn)
{
	*pn = data[0] + data[1] * 256;
	return data + 2;
}

bool Gene::reverseTranscription(const u8 *data)
{
	if(env == NULL) return false;

	int n;

#if defined(SEND_FITNESS_OVER_NETWORK)
	const char *ff8 = "\xff\xff\xff\xff\xff\xff\xff\xff";
	if(memcmp(data, ff8, 8) == 0)
		dFitness = DBL_MAX;
	else
	{
		i32 mant = ntohl(*(u32 *)data);
		n = ntohl(*(u32 *)(data + 4));
		dFitness = ldexp((double)mant, n - 30);
	}
#else
	dFitness = DBL_MAX;
#endif

	n = data[RNA_OFFSET] + data[RNA_OFFSET + 1] * 256;
	if(n != env->getChromoLength()) return false;
	memcpy(b, data + RNA_OFFSET + 2, n);

#if defined(USE_COLOR_PROFILING)
	const u8 *p = data + n + 2 + RNA_OFFSET;
	n = p[0] + p[1] * 256;
	if(n != env->getSizeOfEnvironmentColoringStr())
		return false;
	memcpy(colors, p + 2, n);
#endif
	return true;
}

bool Gene::read(const char *str)
{
	int len = strlen(str) / 2;
	int i;
	unsigned int temp;
	for(i = 0; i < len && i < getChromoLength(); i++)
	{
		sscanf(str + i * 2, "%02X", &temp);
		b[i] = temp;
	}
	return true;
}

int Gene::index() const
{
	const Gene *start = &(*env)[0];
	return (int)(this - start);
}

bool Gene::operator==(const Gene& gen) const
{
	return !memcmp(b, gen.b, env->getChromoLength());
}

bool Gene::operator<(const Gene& gen) const
{
	if(fitness() < gen.fitness()) return true;
	return false;
}

bool Gene::operator>(const Gene& gen) const
{
	if(fitness() > gen.fitness()) return true;
	return false;
}

//////////////////////////////////////////////////////////////////////////////
// GAObserver
//
GAObserver::GAObserver() : env(NULL)
{
}

GAObserver::~GAObserver()
{
	if(env)
		env->deleteObserver(this);
}

//////////////////////////////////////////////////////////////////////////////
// GAObserver related GeneEnvironment methods
//
bool GeneEnvironment::addObserver(GAObserver *observer) const
{
	int i;
	if(observers)
	{
		for(i = 0; observers[i] != NULL; i++)
			if(observers[i] == observer) return true;
	}
	else i = 0;

	GAObserver **temp = new GAObserver *[i + 2];
	temp[i] = observer;
	temp[i + 1] = NULL;
	observer->env = this;
	if(observers)
	{
		memcpy(temp, observers, sizeof(GAObserver*) * i);
		delete[] observers;
	}
	observers = temp;
	return true;
}

bool GeneEnvironment::deleteObserver(GAObserver *observer) const
{
	if(observers)
	{
		int i;
		for(i = 0; observers[i] != NULL; i++)
			if(observers[i] == observer)
			{
				observers[i]->env = NULL;
				for(; observers[i + 1] != NULL; i++)
					observers[i] = observers[i + 1];
				observers[i] = 0;
				return true;
			}
	}
	return true;
}

bool GeneEnvironment::notify(int stage, constgene_pointer *array, int size) const
{
	if(observers)
	{
		int i;
		bool f = false;
		beginUpdate();
		for(i = 0; observers[i] != NULL; i++)
			f |= observers[i]->catchEvent(this, stage, array, size);
		endUpdate();
		return true;
	}
	return false;
}


//////////////////////////////////////////////////////////////////////////////
// GeneEnvironment
//
GeneEnvironment::GeneEnvironment()
{
	primary = secondary = immigrants = elite = NULL;
	afNewElite = NULL;
	ap = apGene = NULL;
	work = NULL;
	mr_tbl = NULL;
	observers = NULL;

	adMinimum = NULL;

	cleanupEnvironment();
}

GeneEnvironment::~GeneEnvironment()
{
	if(observers)
	{
		int i;
		for(i = 0; observers[i] != NULL; i++)
			deleteObserver(observers[i]);
		delete[] observers;
	}
	deleteInstance();
}

void GeneEnvironment::cleanupEnvironment()
{
	tt = 0;
	tag = NULL;
	nId = 0;
	nEnvironments = 1;
	nUpdateCount = 0;
	fPrepareSortedList = false;
	nChromoLength = 16;
	nElite = 10;
	count = 100;
	nCrossoverPoint = 1;
	dCrossoverRate = 0.5;
	dMutationRatePrev = -2.0;
	dMutationRate = -1.0; // not set
	nEmigrants = nMigWPtr = nMigRPtr = 0;
	nMigInterval = 10;
	fEliteMigration = false;

	scalingMethod  = smNormal;
	mutationMethod = mmNormal;
	scalingFunc    = defaultScalingFunc;

	nScalingWinSize = 5;

	fFirstGeneration = true;
	fEnvEstablished = true;
	deleteInstance();
}

// make a copy of env (internally called)
void GeneEnvironment::copy(const GeneEnvironment& env)
{
	nUpdateCount       = env.nUpdateCount;
	fPrepareSortedList = false;

	nScalingWinPos   = env.nScalingWinPos;
	nScalingWinSize  = env.nScalingWinSize;

	nId                = env.nId;
	nEnvironments      = env.nEnvironments;

	tt                 = env.tt;
	nMigInterval     = env.nMigInterval;
	nEmigrants         = env.nEmigrants;
	nMigWPtr           = env.nMigWPtr;
	nMigRPtr           = env.nMigRPtr;
	nMaxImmigrants     = env.nMaxImmigrants;
	fFirstGeneration   = env.fFirstGeneration;
	nChromoLengthBit   = env.nChromoLengthBit;
	nChromoLength      = env.nChromoLength;
	count              = env.count;
	nElite             = env.nElite;
	nCrossoverPoint    = env.nCrossoverPoint;
	dCrossoverRate     = env.dCrossoverRate;
	dMutationRate      = env.dMutationRate;
	tag                = env.tag;
	fEliteMigration    = env.fEliteMigration;

	scalingMethod      = env.scalingMethod;
	mutationMethod     = env.mutationMethod;
	scalingFunc        = env.scalingFunc;

	if(!env.fEnvEstablished) return;
	allocateMemory();
	fEnvEstablished    = env.fEnvEstablished;

	int i;
	for(i = 0; i < env.count; i++)
	{
		primary[i] = env.primary[i];
		primary[i].naturalize(*this);
		secondary[i] = env.secondary[i];
		secondary[i].naturalize(*this);
	}

	for(i = nMigRPtr; i < nMigWPtr; i++)
	{
		immigrants[i] = env.immigrants[i];
		immigrants[i].naturalize(*this);
	}

	memcpy(adMinimum, env.adMinimum, nScalingWinSize * sizeof(double));
	memcpy(mr_tbl, env.mr_tbl, (nChromoLengthBit + 1) * sizeof(double));
}

// Copy Constructor
GeneEnvironment::GeneEnvironment(const GeneEnvironment& env)
{
	copy(env);
}

// substitute operator
GeneEnvironment& GeneEnvironment::operator=(const GeneEnvironment& env)
{
	copy(env);
	return *this;
}

void GeneEnvironment::allocateMemory()
{
	deleteInstance();

	// bare genes
	primary   = new Gene[count];
	secondary = new Gene[count];
	elite     = new Gene[nElite];

	// elite flag
	afNewElite = new bool[nElite];

	nMaxImmigrants = count / 2 + 10; // 0% ~ 50%
	immigrants = new Gene[nMaxImmigrants]; // it may be enough...

	// work area
	apGene    = new constgene_pointer[count + 1]; // for operations
	ap        = new constgene_pointer[count + 1]; // for sort
	work      = new double[count + 2];
	mr_tbl    = new double[nChromoLengthBit + 2];

	adMinimum = new double[nScalingWinSize];
}

void GeneEnvironment::deleteInstance()
{
	if(!fEnvEstablished)
		return;

	if(primary)    {delete[] primary;    primary = NULL;}
	if(secondary)  {delete[] secondary;  secondary = NULL;}
	if(elite)      {delete[] elite;      elite = NULL;}
	if(afNewElite) {delete[] afNewElite; afNewElite = NULL;}
	if(immigrants) {delete[] immigrants; immigrants = NULL;}
	if(apGene)     {delete[] apGene;     apGene = NULL;}
	if(ap)         {delete[] ap;         ap = NULL;}
	if(work)       {delete[] work;       work = NULL;}
	if(mr_tbl)     {delete[] mr_tbl;     mr_tbl = NULL;}

	if(adMinimum)  {delete[] adMinimum;  adMinimum = NULL;}

	fEnvEstablished = false;
	dMutationRatePrev = -2.0;
}

// confirm parameter settings and realize the environment
// now establishEnvironment is only a stub to call resetEnvironment
bool GeneEnvironment::establishEnvironment()
{
	resetEnvironment();
	return true;
}

void GeneEnvironment::resetEnvironment()
{
	if(!fEnvEstablished)
	{
		allocateMemory();

		// default mutation rate 1/L
		if(dMutationRate < 0)
			dMutationRate = 1.0 / nChromoLengthBit / 8.0;

		reculcMutationTable();
	}

	int i;
	// Give citizenship of this environment to them
	for(i = 0; i < count; i++)
	{
		primary[i].acquireCitizenship(*this);
		primary[i].randomize();
		secondary[i].acquireCitizenship(*this);
		secondary[i].randomize();

		// initialize pointer
		// we suppose that apGene is always initialized
		// before calling prepareSortedList
		apGene[i] = &primary[i];
	}

	for(i = 0; i < nMaxImmigrants; i++)
	{
		immigrants[i].acquireCitizenship(*this);
		immigrants[i].randomize();
	}

	// scaling window method work area
	nScalingWinPos = 0;
	for(i = 0; i < nScalingWinSize; i++)
		adMinimum[i] = DBL_MAX;

	fFirstGeneration = true;
	nMigWPtr = nMigRPtr = 0;
	tt = 0;
	fPrepareSortedList = false;
	fEnvEstablished = true;
}

void GeneEnvironment::setMutationRate(double d)
{
	dMutationRate = d;
	if(mr_tbl) reculcMutationTable();
}

void GeneEnvironment::reculcMutationTable()
{
	if(dMutationRate == dMutationRatePrev)
		return;

	int i, j;
	// Mutation Rate to Number of Mutate points Conversion Table
	mr_tbl[0] = 0;
	for(i = 0; i <= nChromoLengthBit; i++)
	{
		// you may think it is not effecient,
		// but it's the way to calculate accurately.
		// do you imagine how may bits is required to calcurate 0.01^100 ?
		double d = 1;
		for(j = 1; j <= i; j++) d *= dMutationRate;
		for(; j <= nChromoLengthBit; j++) d *= (1 - dMutationRate) * j;
		for(j = 1; j <= nChromoLengthBit - i; j++) d /= j;
		mr_tbl[i + 1] = mr_tbl[i] + d;
	}
	dMutationRatePrev = dMutationRate;
}

bool GeneEnvironment::setNumEnvironments(int n)
{
	if(nUpdateCount) return false;
	nEnvironments = n;
	//deleteInstance();
	return true;
}

bool GeneEnvironment::setEnvironmentId(int n)
{
	if(nUpdateCount) return false;
	nId = n;
	//deleteInstance();
	return true;
}

bool GeneEnvironment::setChromoLengthByBit(int n)
{
	if(nUpdateCount) return false;
	nChromoLengthBit = n; nChromoLength = (n + 7) / 8;
	setMutationRate((double)1.0 / nChromoLengthBit);
	deleteInstance();
	return true;
}

bool GeneEnvironment::setPopulationSize(int n)
{
	if(nUpdateCount) return false;
	count = n;
	deleteInstance();
	return true;
}

int GeneEnvironment::getNumMutationPoints() const
{
	return bin_search(mr_tbl, genrand(), 0, nChromoLengthBit);
}

bool GeneEnvironment::setScalingWindowSize(int n)
{
	if(nUpdateCount) return false;
	nScalingWinSize = n;
	deleteInstance();
	return true;
}
void GeneEnvironment::storeValue(double d)
{
	adMinimum[nScalingWinPos++] = d;
	if(nScalingWinPos >= nScalingWinSize) nScalingWinPos = 0;
}

double GeneEnvironment::loadMinimumValue() const
{
	int i;
	double min = DBL_MAX;
	for(i = 0; i < nScalingWinSize; i++)
		if(min > adMinimum[i]) min = adMinimum[i];
	return min;
}

// insertion sort
static void inssort(const Gene **apGene, int first, int end)
{
	int i, j;
	const Gene *temp;

	for(i = 1 + first; i < end; i++)
	{
		temp = apGene[i];
		for(j = i - 1;
			j >= first && apGene[j]->fitness() < temp->fitness(); j--)
			apGene[j + 1] = apGene[j];
		apGene[j + 1] = temp;
	}
}

// quick sort (decremental direction)
static void qsort(const Gene **apGene, int first, int end)
{
	// quick sort is slower than insertion sort when range is too short.
	if(end - first <= 10)
		inssort(apGene, first, end);

	int i, j;
	double val;
	const Gene *temp;

	val = apGene[(first + end) / 2]->fitness();
	i = first; j = end;
	for(;;)
	{
		while(apGene[i]->fitness() > val) i++;
		while(val > apGene[j]->fitness()) j--;
		if(i >= j) break;

		temp = apGene[i]; apGene[i] = apGene[j]; apGene[j] = temp;
		i++; j--;
	}
	if(first < i - 1) qsort(apGene, first, i - 1);
	if(j + 1 < end)   qsort(apGene, j + 1, end);
}

// getFitness: default fitness evaluator
// Gene::fitness is somewhat faster than GeneEnvironment::getFitness
// because of cache.
// But if fitness function easy to evaluate realtimely,
// you may use getFitness for the purpose of speed up
// GREATER(GOOD) <------> LESS(BAD)
double GeneEnvironment::getFitness(const Gene& gen) const
{
	int i, n;
	for(i = n = 0; i < nChromoLength; i++)
		n += gen.getChromoPtr()[i];
	return (double)(255 * nChromoLength - n);
}

// beginUpdate declares starting point of updating process,
// it prevents from calling quick sort every time and preserve processor power.
// you must call endUpdate in the end of updating process.
// [NOTE] YOU CAN CALL beginUpdate/endUpdate RECURSIVELY.
void GeneEnvironment::beginUpdate() const
{
	nUpdateCount++;
}

void GeneEnvironment::endUpdate() const
{
	nUpdateCount--;
	if(nUpdateCount < 0)
	{
		// do nothing
	}
}

void GeneEnvironment::prepareSortedList() const
{
	if(nUpdateCount) return;
	if(fPrepareSortedList)
		return;

	int i;
	for(i = 0; i < count; i++)
		ap[i] = apGene[i];

	qsort(ap, 0, count - 1);
	fPrepareSortedList = true;
}

const Gene& GeneEnvironment::sortedGene(int n) const
{
	prepareSortedList();
	return *ap[n];
}

// when you call this function(not const ver.) without beginUpdate() and
// endUpdate(), this function calls qsort every time.
// you supposed to call this method between beginUpdate() and endUpdate(),
// or only refer to data by calling sortedGene(const ver.)
Gene& GeneEnvironment::sortedGene(int n)
{
	prepareSortedList();
	fPrepareSortedList = false;
	return *const_cast<Gene *>(ap[n]); // sorry...
}

// this method do one step
// user must call this method every generation.
double GeneEnvironment::nextGeneration()
{
	int i, j;
	bool fShutdown = false;

	// for the people who don't call establishEnvironment()...
	if(!fEnvEstablished)
		establishEnvironment();

	//
	// -------------- SELECTION -------------------
	//
	fShutdown |= notify(0, apGene, count);

	// e_count is an even value to keep consistency of crossover
	int e_count = (count + 1) & (-1 - 1);

	// Pseudo Rank selection may behave different if population size is
	// too small. We knows count = 6, there's a certain difference which is
	// visible by using normal X-Y graph.
	if(scalingMethod == smByPseudoRank)
	{
		// This scaling method likes smByRank very much,
		// but it may cost less than that one.
		prepareSortedList();
		for(i = 0; i < e_count; i++)
		{
			int n = (int)((double)e_count * (1 - sqrt(genrand())));
			apGene[i] = ap[n];
		}
	}
	else if(scalingMethod == smByRank)
	{
		prepareSortedList();
		double t = (double)e_count * 2 + 1.0;
		double pmx = (double)e_count * (e_count + 1) * 4;
		for(i = 0; i < e_count; i++)
		{
			double px8 = genrand() * pmx;
			int n = (int)((t - sqrt(t * t - px8)) / 2);
			apGene[i] = ap[n];
		}
	}
	else
	{
		// check maximum/minimum fitness
		double d;
		double min_fitness = DBL_MAX;
		double max_fitness = -DBL_MAX;
		for(i = 0; i < count; i++)
		{
			d = primary[i].fitness();
			if(d < min_fitness) min_fitness = d;
			else if(d > max_fitness) max_fitness = d;
		}

		if(scalingMethod == smByScalingWin)
		{
			storeValue(min_fitness);
			min_fitness = loadMinimumValue();
		}

		// create rouletting table
		for(i = 0, work[0] = 0.0; i < count; i++)
			work[i + 1] = work[i]
				+ scalingFunc(primary[i].fitness(), max_fitness, min_fitness);

		// if(work[count] == 0)
		if(work[count] > DBL_EPSILON)
		{
			// play roulette to select chromosomes to survive
			// the number of chromosomes survive here is (count - nElite).
			// 1 odd instance prevents this program from crush caused by
			// not even temporary population.
			for(i = 0; i < e_count; i++)
			{
				int n = bin_search(work, genrand() * work[count], 0, count);
				apGene[i] = &primary[n];
			}
		}
		else
		{
			// all chromosomes has same fitness value,
			// simply do random sampling.
			// this also prevents binary search from looping.
			for(i = 0; i < e_count; i++)
				apGene[i] = &primary[genrand_int(count)];
		}
	}

	//
	// -------------- MIGRATION -------------------
	//
	fPrepareSortedList = false;
	fShutdown |= notify(1, apGene, e_count);

	bool f;
	if(tt % nMigInterval)
	{
		if(fEliteMigration)
		{
			prepareSortedList();
			f = migrationProcedure(ap, nEmigrants);
		}
		else
			f = migrationProcedure(apGene, nEmigrants);
	}
	else
		// each generation, call migrationProcedure whether if you
		// you want or not.
		f = migrationProcedure(NULL, 0);

	// someone may want to shutdown now...
	fShutdown |= f;

#if defined(_MT)
	lockVectorGeneInstance();
#endif
	// the immigrants replaces bad-fitness chromosomes
	for(j = 0, i = nMigRPtr; i != nMigWPtr; i = (i + 1) % nMaxImmigrants, j++)
	{
		apGene[count - j - 1] = &immigrants[i];
		immigrants[i].naturalize(*this);
	}

#if defined(_MT)
	releaseVectorGeneInstance();
#endif

	//
	// -------------- CROSSOVER -------------------
	//
	fPrepareSortedList = false;
	fShutdown |= notify(2, apGene, e_count);

	// nHalfPop crossover-pairing forces each immigrant to crossover with
	// a selected individual of this island in this generation.
	int nHalfPop = (count + 1) / 2;
	for(i = 0; i < nHalfPop; i++)
	{
		const Gene &a = *apGene[i];
		const Gene &b = *apGene[i + nHalfPop];
		if(genrand() < dCrossoverRate)
		{
			// crossover, only children can survive this generation
			a.crossover(b, secondary[i], secondary[i + nHalfPop]);
		}
		else
		{
			// do nothing , simply inherit same Chromosomes
			secondary[i] = a;
			secondary[i + nHalfPop] = b;
		}
	}

#if defined(_MT)
	lockVectorGeneInstance();
	nMigRPtr = nMigWPtr;
	releaseVectorGeneInstance();
#else
	nMigRPtr = nMigWPtr;
#endif

	//
	// -------------- MUTATION -------------------
	//
	for(i = 0; i < count; i++)
		apGene[i] = &secondary[i];
	fPrepareSortedList = false;
	fShutdown |= notify(3, apGene, count);

	if(mutationMethod == mmGenesisCompatible && nChromoLengthBit <= count)
	{
		// GENESIS compatible mutation
		int nPt = genrand_int(nChromoLengthBit);
		for(i = 0; i < count; i++, nPt = (nPt + 1) % nChromoLengthBit)
		{
			u8 *p = secondary[i].getChromoPtr();
			p[nPt >> 3] ^= (128 >> (nPt & 7));
		}
	}
	else
	{
		for(i = 0; i < count; i++)
			secondary[i].mutate();
	}

	//
	// -------------- ELITE PROCESSING -------------------
	//
	fPrepareSortedList = false;
	fShutdown |= notify(4, apGene, count);

	if(!fFirstGeneration)
	{
		// some chromosomes are destroyed by elite instance !!
		// but anything very serious can't be caused by this process maybe...
		for(i = 0; i < nElite; i++)
			secondary[genrand_int(count)] = elite[i];

		// sort to extract (and store) elite chromosomes
		prepareSortedList();

		/*// preserve elite chromosomes
		for(i = 0; i < nElite; i++)
			elite[i] = *ap[i];*/

		// elite mergesort
		Gene *temp = new Gene[nElite];
		int k;
		for(i = j = k = 0; i < nElite && j < nElite && k < nElite; i++)
		{
			if(*ap[j] > elite[k])
				temp[i] = *ap[j++], afNewElite[i] = true;
			else
				temp[i].exchange(elite[k++]), afNewElite[i] = false;
		}
		if(k == nElite)
			for(; i < nElite; i++, j++)
				temp[i] = *ap[j], afNewElite[i] = true;
		else
			for(; i < nElite; i++, k++)
				temp[i].exchange(elite[k]), afNewElite[i] = false;
		delete[] elite;
		elite = temp;
	}
	else
	{
		fFirstGeneration = false;

		// sort to extract (and store) elite chromosomes
		prepareSortedList();

		// preserve elite chromosomes
		for(i = 0; i < nElite; i++)
			elite[i] = *ap[i];
	}

	// flipping
	Gene *temp = primary;
	primary = secondary; secondary = temp;

	fPrepareSortedList = true;
	tt++;

	// if shutdown, return NG_SHUTDOWN_SIGNAL
	if(fShutdown)
		return NG_SHUTDOWN_SIGNAL;

	// return best fitness
	return ap[0]->fitness();
}

// for the purpose of Transcription of Chromosomes over the network, etc...
// This method provides VECTOR of the chromosomes.
// YOU MUST CALL THIS FUNCTION BETWEEN lockVectorGeneInstance AND
// releaseVectorGeneInstance IN MULTI-THREADED ENVIRONMENT !!
Gene& GeneEnvironment::getVectorGeneInstance()
{
	// this system always provide many vector gene instance,
	// and enough in normal situation,
	// but sometimes this implementation causes serious system confliction.
	// A way to prevent this is very easy, not want to get many genes migrated.
	Gene& imm = immigrants[nMigWPtr];
	imm.clearFitnessCache();
	nMigWPtr = (nMigWPtr + 1) % nMaxImmigrants;
	return imm;
}

bool GeneEnvironment::isNewElite(const Gene& gen) const
{
	if(&gen >= elite && &gen < elite + nElite)
		return afNewElite[&gen - elite];
	return false;
}

bool GeneEnvironment::migrationProcedure(const constgene_pointer apEmigrants[], int nEmigrants)
{
	// this implement provides migrationProcedure for the sake of SGA users,
	// but nothing to do... (Simple GA)
	// DEGA projects must override this method.
	//
	// if you want to terminate GA execusion, return true
	return false;
}

istream& operator>>(istream& in, Gene& gen)
{
	char buf[256];
	in >> buf;

	gen.read(buf);
	return in;
}

// extractHexStr string version
string Gene::extractHexStr() const
{
	string str;
	int size = getChromoLength() * 2 + 1;
	char *chromo = new char[size];

	extractHexStr(chromo, size);
	str = chromo;
	delete[] chromo;
	return str;
}

// GeneEnvironment::getString only calls extractHexStr
string GeneEnvironment::getString(const Gene& gen) const
{
	return gen.extractHexStr();
}

// Gene::getString only calls GeneEnvironment::getString
string Gene::getString() const
{
	string str;
	str = env->getString(*this);
	return str;
}

// ostream operator
// it enables the following form:
// cout << gen << endl;
ostream& operator<<(ostream& out, const Gene& gen)
{
	out << gen.getString();
	return out;
}

// extractColoringStr string version
#if defined(USE_COLOR_PROFILING)
string Gene::extractColoringStr() const
{
	string str;
	int size = getColorLength() + 1;
	char *buffer = new char[size];
	extractColoringStr(buffer, size);
	str = buffer;
	delete[] buffer;
	return str;
}
#endif

bool GAMonitor::catchEvent(const GeneEnvironment *env, int stage, constgene_pointer *array, int popsize)
{
	constgene_pointer *temp = new constgene_pointer[popsize];
	memcpy(temp, array, sizeof(constgene_pointer) * popsize);
	qsort(temp, 0, popsize - 1);
	bool f = monitor(env, stage, temp, popsize);
	delete[] temp;
	return f;
}
