//--------------------------------------------------------------------------
// dega.h
// GAPPA Parallel Distributed Environment Genetic Algorithm Library
// DEGA main logic
//--------------------------------------------------------------------------
// Copyright (C) 2000 Takashi Kawasaki, Tomoyuki Hiroyasu and Mitsunori Miki,
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright  notice
//    and this list of conditions.
// 2. Redistributions in binary form must reproduce the above  copyright notice
//    and this list of conditions in the documentation and/or other materials
//    provided with the distribution.
// 3. All advertising materials mentioning features or use of this software
//    must display the following acknowledgement:
//
//    This product includes software developed by Takashi Kawasaki,
//    Tomoyuki Hiroyasu and Mitsunori Miki except the functions of sgenrand()
//    and genrand().
//
// 4. The names of Takashi Kawasaki, Tomoyuki Hiroyasu and Mitsunori Miki
//    should not be used to endorse or promote products derived from this
//    software without specific prior written permission.

#ifndef _GAPPA_DEGA_LIBRARY_HEADER_
#define _GAPPA_DEGA_LIBRARY_HEADER_

//#define _MT 1
#define OPTIMIZATION_FOR_SPEED 1
//#define USE_COLOR_PROFILING 1
#define FORCE_CHROMO_STRICT_CHECK 1
//#define NO_FITNESS_CACHE 1
//#define DF_LONGER_THAN_8BIT 1
#define USE_REPORTING_INTERFACE 1

// in current(01/07/2000) Win32/Intel Linux architecture,
// double has 52bit mantissa-part and int has 32bits. Because of this
// difference, int can not cover all integer values expressed by double.
// This macro provides you with the solution by double but may result in
// lower performance.
// #define OPTIMIZED_FOR_PRECISION 1

#define SEND_FITNESS_OVER_NETWORK 1

#include <iostream>
#include <string>
#include <cstdio>
#include <cstdlib>
#include <cfloat> // current gcc doesn't support <limits>

#include "sysdep.h"
#include "mtrand.h"

#if defined(_MT)
#include "muthread.h"
#endif

#if defined(SEND_FITNESS_OVER_NETWORK)
#define RNA_OFFSET 8
#else
#define RNA_OFFSET 0
#endif

#define NO_ENVIRONMENT 0
#define MISSING_BEGINUPDATE 1
#define GENE_NOT_FOUND 2
#define MISSMATCH_LENGTH 3
#define GENE_FLAT_FITNESSES 4
#define NO_MORE_VECTOR 5

class GeneEnvironment;

// DesignVariableDomain
// the result value is calcurated by following formula:
//
//  x = X * (openMax - closedMin) / 2^nBit + closedMin
//
//  x : result value  (closedMin <= x < openMax)
//  X : a integer value extracted from chromosome, its bit width is described
//      by nBit field.
typedef struct
{
	int nBit;
	double closedMin;
	double openMax;
}DesignVariableDomain;

class BitAccessorProxy
{
	u8 *p;
	int bit;
public:
	BitAccessorProxy(u8 *b, int pos) : p(b + pos / 8), bit(128 >> pos % 8) {}
	BitAccessorProxy& operator=(bool f)
	{
		if(f) *p |= bit; else *p &= ~bit;
		return *this;
	}

	BitAccessorProxy& operator|=(bool f)
	{
		if(f) *p |= bit;
		return *this;
	}

	BitAccessorProxy& operator&=(bool f)
	{
		if(!f) *p &= ~bit;
		return *this;
	}

	BitAccessorProxy& operator^=(bool f)
	{
		if(f) *p ^= bit;
		return *this;
	}
};

class BitAccessor
{
	int size;
	u8 *b;

public:
	explicit BitAccessor(u8 *data, int n = 0) : b(data), size(n) {}
	inline int length() const {return size;}
	bool operator[](int n) const {return b[n / 8] & (128 >> n % 8);}
	BitAccessorProxy operator[](int n) {return BitAccessorProxy(b, n);}
};

#define NG_SHUTDOWN_SIGNAL (-DBL_MAX)

class Gene
{
#if defined(USE_COLOR_PROFILING)
	u8 *colors;
#endif
	const GeneEnvironment *env; // this informs us where does it exist
	u8 *b;                      // chromosome

#if !defined(NO_FITNESS_CACHE)
	mutable double dFitness;    // fitness cache
#endif

	void construct(const GeneEnvironment& environment);
	void destruct();
	void copy(const Gene& gen);
#if defined(USE_COLOR_PROFILING)
	void setEnvironmentColor();
#endif

public:
	inline const GeneEnvironment *getEnvironment() const {return env;}
	void acquireCitizenship(const GeneEnvironment& environment);
	explicit Gene(GeneEnvironment& environment);
	Gene();
	~Gene();
	Gene(const Gene& gen);
	inline Gene& operator=(const Gene& gen) {copy(gen); return *this;}
	bool naturalize(const GeneEnvironment& environment);

#if defined(USE_COLOR_PROFILING)
	const u8 *getColor() const {return colors;}
	inline int getColorLength() const;
#endif

#if !defined(NO_FITNESS_CACHE)
	inline void clearFitnessCache()
#if defined(__BORLANDC__)
	{dFitness = std::DBL_MAX;}
#else
	{dFitness = DBL_MAX;}
#endif
#else
	inline void clearFitnessCache() { /* do nothing */ }
#endif
	void randomize();
	void mutate();
	bool exchange(Gene& gen);
	void crossover(const Gene& partner, Gene& child1, Gene& child2) const;
	double fitness() const;

	inline int getChromoLength() const;
	inline int getChromoLengthByBit() const;
	const u8& operator[](int n) const {return b[n];}

	inline u8 *getChromoPtr() {clearFitnessCache(); return b;}
	u8& operator[](int n) {clearFitnessCache(); return b[n];}

	inline const u8 *getChromoPtr() const {return b;}

	// Design Variable cracking/construction functions
	// integer version
	bool extractDesignVariables(int anDesignVariable[],
		const int *anBitSize, int nDesignVariable,
		bool fBalanced = true, bool fUseGrayCode = false) const;
	bool buildDesignVariables(const int anDesignVariable[],
		const int *anBitSize, int nDesignVariable,
		bool fBalanced = true, bool fUseGrayCode = false);

#if defined(HAVE_64BIT_TYPE)
	bool extractDesignVariables(i64 anDesignVariable[],
		const int *anBitSize, int nDesignVariable,
		bool fBalanced = true, bool fUseGrayCode = false) const;
	bool buildDesignVariables(const i64 anDesignVariable[],
		const int *anBitSize, int nDesignVariable,
		bool fBalanced = true, bool fUseGrayCode = false);
#endif // HAVE_64BIT_TYPE

	// double version
	// by half opened domain : DesignVariableDomain
	bool extractDesignVariables(double adDesignVariable[],
		const DesignVariableDomain *dvd, int nDesignVariable,
		bool fBalanced = true, bool fUseGrayCode = false) const;
	bool buildDesignVariables(const double adDesignVariable[],
		const DesignVariableDomain *dvd, int nDesignVariable,
		bool fBalanced = true, bool fUseGrayCode = false);

	const char* extractHexStr(char *buffer, int size) const;
	std::string extractHexStr() const;
#if defined(USE_COLOR_PROFILING)
	const char* extractColoringStr(char *buffer, int size) const;
	std::string extractColoringStr() const;
#endif

	inline int getRnaLength() const;
	const u8* transcription(u8 *buffer) const;
	bool reverseTranscription(const u8 *data);
	bool read(const char *str);
	inline bool read(std::string str) {return read(str.c_str());}

	std::string getString() const;

	inline bool operator()(int n) const {return b[n >> 3] & (128 >> (n & 7));}
	void setBit(int n, bool f);

	const BitAccessor bit_string() const
		{return BitAccessor(b, getChromoLengthByBit());}
	BitAccessor bit_string()
	{
		clearFitnessCache();
		return BitAccessor(b, getChromoLengthByBit());
	}

	int index() const;
	bool operator==(const Gene& gen) const;
	bool operator<(const Gene& gen) const;
	bool operator>(const Gene& gen) const;
	bool operator>=(const Gene& gen) const {return *this > gen || *this == gen;}
	bool operator<=(const Gene& gen) const {return *this < gen || *this == gen;}
	bool operator!=(const Gene& gen) const {return !(*this == gen);}
};

typedef const Gene* constgene_pointer;

// GA Debugging/Monitoring Interface Object
class GAObserver
{
	friend GeneEnvironment;
	const GeneEnvironment *env;

protected:
	virtual bool catchEvent(const GeneEnvironment *env, int stage, constgene_pointer *array, int popsize) = 0;

public:
	GAObserver();
	virtual ~GAObserver();
};

typedef enum {smNormal = 0, smByScalingWin, smByRank, smByPseudoRank} ScalingMethod;
typedef enum {mmNormal = 0, mmGenesisCompatible} MutationMethod;
typedef double (*ScalingFunction)(double fitness, double max, double min);
double defaultScalingFunc(double fitness, double max, double min);

// Environment Object
class GeneEnvironment
{
	int nId;                         // one of [0 ... nEnvironment - 1]
	int nEnvironments;               // the number of Environments

	int nEmigrants;
	int nMigWPtr, nMigRPtr;          // out, in
	int nMaxImmigrants;              // Immigrants capacity
	int nMigInterval;

	bool fEliteMigration;

	int tt;
	bool fFirstGeneration;
	bool fEnvEstablished;
	int nChromoLengthBit;
	int nChromoLength;     // chromosome length in bytes, not GTYPE
	int count;             // population capacity of this environment
	int nElite;            // the number of elites to survive
	int nCrossoverPoint;   // the number of cross over points
	double dCrossoverRate; // cross over rate
	double dMutationRate;  // mutation rate
	double dMutationRatePrev;

	int nScalingWinPos;              // current position in the buffer
	int nScalingWinSize;

	Gene *primary;                     // Array of Gene #1
	Gene *secondary;                   //               #2
	Gene *elite;
	bool *afNewElite;

	Gene *immigrants;                  // international hotel for immigrants
	constgene_pointer *apGene;         // work area for nextGeneration
	mutable constgene_pointer *ap;     // work area for sorting
	double *work;                      // fitness work area
	double *mr_tbl;                    // mutation prob. distribution table

	mutable GAObserver **observers;

	mutable int nUpdateCount;        // lock count for smart updating
	mutable bool fPrepareSortedList; // whether sorted list is prepared or not

	double *adMinimum;               // work area for Scaling Window method

	void prepareSortedList() const;
	void deleteInstance();
	void copy(const GeneEnvironment& env);

	void storeValue(double d);
	double loadMinimumValue() const;

	void allocateMemory();
	void reculcMutationTable();

#if defined(_MT)
	mutable muMutex mutexImm;
#endif

	ScalingMethod scalingMethod;
	MutationMethod mutationMethod;
	ScalingFunction scalingFunc;

	bool notify(int stage, constgene_pointer *array, int size) const;

public:
	void *tag; // additional variable for communication

	void clearSortedList() const {fPrepareSortedList = false;}

	bool addObserver(GAObserver *observer) const;
	bool deleteObserver(GAObserver *observer) const;

#if defined(USE_COLOR_PROFILING)
	inline int getRnaLength() const
		{return 4 + nChromoLength + nEnvironments + RNA_OFFSET;}
#else
	inline int getRnaLength() const
		{return 2 + nChromoLength + RNA_OFFSET;}
#endif

	inline int getNumEnvironments() const {return nEnvironments;}
	inline int getEnvironmentId() const {return nId;}
	inline void setNumEmigrants(int n) {nEmigrants = n;}
	inline int getNumEmigrants() const {return nEmigrants;}
	inline void setEmigrationRate(double d)
		{nEmigrants = (int)(d * count + 0.5);} // fixed:19991115
	inline double getEmigrationRate() const
		{return (double)nEmigrants / count;}
	inline void resetTimer() {tt = 0;}
	inline int getTimer() const {return tt;}
	inline void setMigrationTerm(int n) {nMigInterval = n;}
	inline int getMigrationTerm() const {return nMigInterval;}
	inline void setMigrationInterval(int n) {nMigInterval = n;}
	inline int getMigrationInterval() const {return nMigInterval;}

#if defined(USE_COLOR_PROFILING)
	inline int getSizeOfEnvironmentColoringStr() const {return nEnvironments;}
#endif

	inline int getNumMutationPoints() const;
	inline void setRandomeSeed(int n) {srand(n);}
	inline int getChromoLengthByBit() const {return nChromoLengthBit;}
	inline int getChromoLength() const {return nChromoLength;}

	// these methods only can be called within initializing
	bool setNumEnvironments(int n);
	bool setEnvironmentId(int n);
	bool setChromoLengthByBit(int n);
	inline bool setChromoLength(int n) { return setChromoLengthByBit(n * 8); }
	bool setPopulationSize(int n);

	void setMutationRate(double d);
	inline double getMutationRate() const {return dMutationRate;}
	inline void setCrossoverRate(double d) {dCrossoverRate = d;}
	inline double getCrossoverRate() const {return dCrossoverRate;}
	inline void forceUniformCrossover() {nCrossoverPoint = 0;}
	inline void setNumCrossoverPoints(int n) {nCrossoverPoint = n;}
	inline int getNumCrossoverPoints() const {return nCrossoverPoint;}
	inline void setNumEliteSurvive(int n) {nElite = n;}
	inline int getNumEliteSurvive() const {return nElite;}
	inline int getPopulationSize() const {return count;}
	inline void enableEliteMigration(bool f) {fEliteMigration = f;}
	inline bool enableEliteMigration() {return fEliteMigration;}

	inline ScalingMethod getScalingMethod() const {return scalingMethod;}
	inline void setScalingMethod(ScalingMethod sm) {scalingMethod = sm;}
	inline MutationMethod getMutationMethod() const {return mutationMethod;}
	inline void setMutationMethod(MutationMethod mm) {mutationMethod = mm;}
	inline ScalingFunction getScalingFunction() const {return scalingFunc;}
	inline void setScalingFunction(ScalingFunction sf) {scalingFunc = sf;}

	bool setScalingWindowSize(int n);
	inline int getScalingWindowSize() const {return nScalingWinSize;}

	GeneEnvironment();
	virtual ~GeneEnvironment();
	GeneEnvironment(const GeneEnvironment& env);
	GeneEnvironment& operator=(const GeneEnvironment& env);

	bool establishEnvironment();
	void cleanupEnvironment();
	virtual void resetEnvironment();

	const Gene& operator[](int n) const {return primary[n];}
	Gene& operator[](int n) {return primary[n];}

	inline const Gene *getEliteArray() const {return elite;}

	// [CAUTION]
	// this functions may have serious effect on nextGeneration processing
	inline Gene *getEliteArray() {return elite;}

	// this function identifies whether the instance is the new elite
	// or not. The new elite is defined as elite instance which has just
	// been registered as elite in last one generation.
	bool isNewElite(const Gene& gen) const;

	// please override this method
	virtual double getFitness(const Gene& gen) const;

	// if something goes to shutdown, this function returns false
	double nextGeneration();

	virtual void shutdown() {}

	// exchage method
	Gene& getVectorGeneInstance();
	virtual bool migrationProcedure(const constgene_pointer apEmigrants[], int nEmigrants);

#if defined(_MT)
	inline void lockVectorGeneInstance() const {mutexImm.lock();}
	inline void releaseVectorGeneInstance() const {mutexImm.unlock();}
#endif

	const Gene& sortedGene(int n) const;
	Gene& sortedGene(int n);
	inline const Gene& operator()(int n) const {return sortedGene(n);}
	inline Gene& operator()(int n) {return sortedGene(n);}

	void beginUpdate() const;
	void endUpdate() const;

	virtual std::string getString(const Gene& gen) const;
};

const u8 *extractRnaInformation(const u8 *data, int *pn);

std::ostream& operator<<(std::ostream& out, const Gene& gen);
std::istream& operator>>(std::istream& in, Gene& gen);

class GAMonitor : public GAObserver
{
protected:
	bool catchEvent(const GeneEnvironment *env, int stage, constgene_pointer *array, int popsize);
public:
	GAMonitor() : GAObserver() {}
	virtual ~GAMonitor() {}
	virtual bool monitor(const GeneEnvironment *env, int stage, constgene_pointer *array, int popsize) = 0;
};

#endif // _GAPPA_DEGA_LIBRARY_HEADER_
