//--------------------------------------------------------------------------
// gafuncs.cpp
// GAPPA Parallel Distributed Environment Genetic Algorithm Library
// Welknown functions implementation
//--------------------------------------------------------------------------
// Copyright (C) 2000 Takashi Kawasaki, Tomoyuki Hiroyasu and Mitsunori Miki,
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright  notice
//    and this list of conditions.
// 2. Redistributions in binary form must reproduce the above  copyright notice
//    and this list of conditions in the documentation and/or other materials
//    provided with the distribution.
// 3. All advertising materials mentioning features or use of this software
//    must display the following acknowledgement:
//
//    This product includes software developed by Takashi Kawasaki,
//    Tomoyuki Hiroyasu and Mitsunori Miki except the functions of sgenrand()
//    and genrand().
//
// 4. The names of Takashi Kawasaki, Tomoyuki Hiroyasu and Mitsunori Miki
//    should not be used to endorse or promote products derived from this
//    software without specific prior written permission.

#include <cmath>
#include "gafuncs.h"

// This module retains many functions usually used by GA reserchers and so on
// and have welknown solutions.
// Since GAPPA only supports maximize problems and for the easiness of use,
// functions below this line may intentionally be modified and/or changed.
// Good luck.

//----------------------------------------------------------------------------
// Rosenbrock [-2.048 2.048) : 0.0
double rosenbrock(const double *x, int n)
{
	int i;

	double d = 0;
	for(i = 0; i < n ; i += 2)
	{
		double p = x[i] - x[i + 1] * x[i + 1];
		double q = x[i + 1] - 1;
		d += p * p * 100 + q * q;
	}
	return -d;
}

//----------------------------------------------------------------------------
// Rastrigin [-5.12,5.12) : 0.0
double rastrigin(const double *x, int n)
{
	int i;
	double d;
	for(i = 0, d = 0; i < n; i++)
		d += x[i] * x[i] - 10 * cos(2 * 3.14159 * x[i]);
	return -(10 * n + d);

}

//----------------------------------------------------------------------------
// Schwefel [-512 512) : 0.0
// this version of schwefel converges to 0.0, because of hard-coding optimal.
double schwefel(const double *x, int n)
{
	// if your platform has none IEEE formatted double, value under this line
	// must be changed or it may effect badly.
	// I know Win32(Intel) and Linux(i386) commonly have IEEE double format.
	static const u32 sol[] = {0xb9e8be9c, 0x4043b5ab};

	int i;
	double d;
	for(i = 0, d = 0; i < n; i++)
		d += x[i] * sin(sqrt(fabs(x[i])));
	return d; //  *(double *)sol * n / 10;
}

//----------------------------------------------------------------------------
// Griewank [-512 512) : 0.0
double griewank(const double *x, int n)
{
	int i;
	double d, e;
	for(i = 0, d = e = 1; i < n; i++){
		d += x[i] * x[i] / 4000;
		e *= cos(x[i] / sqrt(i + 1));
	}
	return -(d - e);
}

//----------------------------------------------------------------------------
// F4: Ridge [-64 64) : 0.0
double ridge(const double *x, int n)
{
	int i, j;
	double d = 0;
	for(i = 0; i < n; i++)
	{
		double p = 0;
		for(j = 0; j < i; j++)
			p += x[j];
		d += p * p;
	}
	return -d;
}

//----------------------------------------------------------------------------
// Onemax problem
// Its optimal solution is obtained when each bit is 1.
double onemax(const BitAccessor& bits)
{
	int i, score;
	for(i = score = 0; i < bits.length(); i++)
		if(bits[i]) score++;
	return (double)score;
}
