//--------------------------------------------------------------------------
// FILENAME    : inetbase.c
// DESCRIPTION : base level socket manipulation library
// Programmed by T.Kawasaki
//--------------------------------------------------------------------------
#include "inetbase.h"

int cleanup_socket_env()
{
#ifdef WINDOWS_CC
	WSACleanup();
#endif
	return 0;
}

int setup_socket_env()
{
#ifdef WINDOWS_CC
	int err;
	WSADATA wsad;
	err = WSAStartup(MAKEWORD(1, 1), &wsad);
	if(err != 0) return -1;
#endif
	return 0;
}

void show_info(const char *caption, SOCKADDR_IN *paddr)
{
	char buf[256];
	u16 port;
	get_hostinfo(buf, &port, paddr);
	printf("%s: HOST %s, PORT %d\n", caption, buf, port);
}

// my host address
u32 get_myhostaddr()
{
	char buf[256];
	gethostname(buf, 256);
	return get_hostaddr(buf); // net order
}

// hostname to IP address
u32 get_hostaddr(const char *hostname)
{
	struct hostent *host;
	u32 naddr = inet_addr(hostname); // net order
	if(naddr != (u32)-1)
		return naddr;

	host = gethostbyname(hostname);
	if(host == NULL)
		return (u32)-1;

	return *(u32 *)host->h_addr_list[0]; // net order
}

u16 split_portinfo(char *hostname)
{
	char *p = strchr(hostname, ':');
	if(p != NULL)
	{
		*p++ = 0;
		return (u16)atoi(p); // host order
	}
	return 0xffff;
}

// make up sockaddr_in struct
int init_hostinfoA(SOCKADDR_IN *addr, u32 ip, u16 port)
{
	memset(addr, 0, sizeof(SOCKADDR_IN));
	addr->sin_family = AF_INET;
	addr->sin_addr.s_addr = ip;
	addr->sin_port = htons((u_short)port);
	return 0;
}

int init_hostinfo(SOCKADDR_IN *addr, const char *hostname, u16 port)
{
	return init_hostinfoA(addr, get_hostaddr(hostname), port);
}

// extract host information from sockaddr_in
void get_hostinfo(char *host, u16 *pport, const SOCKADDR_IN *saddr)
{
	strcpy(host, inet_ntoa(saddr->sin_addr));
	*pport = ntohs(saddr->sin_port);
}

const char *get_hostaddrstr(u32 addr, char *buffer)
{
	static char buf[256];
	struct in_addr ia;
	ia.s_addr = addr;
	if(buffer == NULL) buffer = buf;
	strcpy(buffer, inet_ntoa(ia));
	return buffer;
}

SOCKET s_connect_sockaddr(SOCKADDR_IN *paddr)
{
	SOCKET fd;
	fd = socket(AF_INET, SOCK_STREAM, 0);
	if(fd < 0)
		return INVALID_SOCKET;

	if(connect(fd, (struct sockaddr *)paddr, sizeof(SOCKADDR_IN)) < 0)
	{
		closesocket(fd);
		return INVALID_SOCKET;
	}
	return fd;
}

SOCKET s_connectA(u32 ip, u16 port, SOCKADDR_IN *paddr)
{
	SOCKADDR_IN addr;

	if(paddr == NULL)
		paddr = &addr;
	init_hostinfoA(paddr, ip, port);
	return s_connect_sockaddr(paddr);
}

SOCKET s_connect(const char *hostname, u16 port, SOCKADDR_IN *paddr)
{
	return s_connectA(get_hostaddr(hostname), port, paddr);
}

SOCKET s_connect_host(const char *hostname, SOCKADDR_IN *paddr)
{
	char buf[256];
	u16 port;
	strcpy(buf, hostname);
	port = split_portinfo(buf);
	if(port == 0xffff) return INVALID_SOCKET;
	return s_connect(buf, port, paddr);
}

SOCKET prepare_accept(u16 port)
{
	SOCKADDR_IN addr;
	SOCKET fd;

	fd = socket(AF_INET, SOCK_STREAM, 0);
	if(fd == INVALID_SOCKET) return INVALID_SOCKET;

	init_hostinfoA(&addr, get_myhostaddr(), port);
	if(bind(fd, (struct sockaddr *)&addr, sizeof(SOCKADDR_IN)) < 0)
	{
		closesocket(fd);
		return INVALID_SOCKET;
	}

	listen(fd, SOMAXCONN);
	return fd;
}

SOCKET s_accept(SOCKET fd_accept, SOCKADDR_IN *paddr)
{
	SOCKET fd;
	SOCKADDR_IN addr;
#ifdef WINDOWS_CC
	int n;
#else
	socklen_t n;
#endif

	if(paddr == NULL) paddr = &addr;
	n = sizeof(SOCKADDR_IN);
	fd = accept(fd_accept, (struct sockaddr *)paddr, &n);
	return fd;
}

SOCKET accept_if_any(SOCKET fd_accept, SOCKADDR_IN *paddr)
{
	int fAny;
	s_select(&fd_accept, &fAny, 1, 0);
	if(fAny) return s_accept(fd_accept, paddr);
	paddr = NULL;
	return INVALID_SOCKET;
}

// reliable sending
int s_write(SOCKET fd, const void *pv, int size)
{
	int n;
	for(n = 0; n < size;)
	{
		int t = send(fd, (char *)pv + n, size - n, 0);
		if(t < 0) return t;
		n += t;
	}
	return n;
}

void s_write_i32(SOCKET fd, i32 n)
{
	i32 temp = htonl(n);
	s_write(fd, &temp, sizeof(i32));
}

void s_write_i16(SOCKET fd, i16 n)
{
	i16 temp = htons(n);
	s_write(fd, &temp, sizeof(i16));
}

// reliable receiving
int s_read(SOCKET fd, void *pv, int size)
{
	int n;
	for(n = 0; n < size;)
	{
		int t = recv(fd, (char *)pv + n, size - n, 0);
		if(t < 0)
			return t;
		n += t;
	}
	return n;
}

i32 s_read_i32(SOCKET fd)
{
	i32 temp;
	s_read(fd, &temp, sizeof(i32));
	return ntohl(temp);
}

i16 s_read_i16(SOCKET fd)
{
	i16 temp;
	s_read(fd, &temp, sizeof(i16));
	return ntohs(temp);
}

int sfputs(SOCKET fd, const char *p)
{
	int n, err = -1;
	n = s_write(fd, p, strlen(p));
	if(n < 0) return err;
	err = s_write(fd, "\xD\xA", 2);
	if(err < 0) return err;
	return n + err;
}

int sfgets(SOCKET fd, char *buffer)
{
	int n = 0, err;
	do{
		err = recv(fd, buffer + n, 1, 0);
		if(err < 0) return err;
		n++;
	}while(buffer[n - 1] != 0xA);
	n -= 2;
	buffer[n] = 0;
	return n;
}

int s_select(SOCKET socklist[], int available[], int num, int timeout)
{
	struct timeval tv;
	struct timeval *ptv;
	fd_set fds;
	int i, n;

	if(timeout < 0)
		ptv = NULL;
	else
	{
		ptv = &tv;
		tv.tv_sec = timeout / 1000;
		tv.tv_usec = timeout % 1000;
	}

	FD_ZERO(&fds);
	for(i = 0, n = 0; i < num; i++)
	{
		if(socklist[i] > 0)
		{
			FD_SET(socklist[i], &fds);
			if(n < socklist[i]) n = socklist[i];
		}
	}

	if((n = select(n + 1, &fds, NULL, NULL, ptv)) > 0)
	{
		for(i = 0; i < num; i++)
			if(socklist[i] > 0)
				available[i] = FD_ISSET(socklist[i], &fds);
			else
				available[i] = 0;
		return n;
	}
	return n;
}
