//--------------------------------------------------------------------------
// FILENAME    : inifile.cpp
// GAPPA Parallel Distributed Environment Genetic Algorithm Library
// read/write INI formated file
//--------------------------------------------------------------------------
// Copyright (C) 2000 Takashi Kawasaki, Tomoyuki Hiroyasu and Mitsunori Miki,
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright  notice
//    and this list of conditions.
// 2. Redistributions in binary form must reproduce the above  copyright notice
//    and this list of conditions in the documentation and/or other materials
//    provided with the distribution.
// 3. All advertising materials mentioning features or use of this software
//    must display the following acknowledgement:
//
//    This product includes software developed by Takashi Kawasaki,
//    Tomoyuki Hiroyasu and Mitsunori Miki except the functions of sgenrand()
//    and genrand().
//
// 4. The names of Takashi Kawasaki, Tomoyuki Hiroyasu and Mitsunori Miki
//    should not be used to endorse or promote products derived from this
//    software without specific prior written permission.

#include <cstdio> // this class internally uses standard C I/O library.
#include <cstdlib>
#include <string>

using namespace std;

#include "inifile.h"

int seek_break(FILE *fp)
{
	char buf[2000];
	int bytes_read, pos, re;
	int i;

	pos = ftell(fp);
	bytes_read = 0;
	for(;;)
	{
		re = fread(buf, 1, 2000, fp);
		if(re > 0)
			bytes_read += re;
		else
		{
			fseek(fp, pos, SEEK_SET);
			if(bytes_read)
				return bytes_read;
			return -1;
		}

		for(i = 0; i < re; i++)
			if(buf[i] == 0xA || buf[i] == 0xD)
			{
				fseek(fp, pos, SEEK_SET);
				return bytes_read - re + i;
			}
	}
}

void free_freadline(char *p)
{
	free(p);
}

char *freadline(FILE *fp)
{
	int n;
	char *buffer, temp;

	n = seek_break(fp);
	if(n < 0) return NULL;

	buffer = new char[n + 1];
	if(buffer == NULL) return NULL;

	fread(buffer, 1, n + 1, fp);
	if(buffer[n] == 0xD) // CR
	{
		if(fread(&temp, 1, 1, fp))
			if(temp != 0xA)
				fseek(fp, -1, SEEK_CUR);
	}
	buffer[n] = 0;
	return buffer;
}

IniFile::IniFile(const char *filename)
{
	count = 0;
	pList = NULL;
	fChanged = false;
	strcpy(fn, filename);

	FILE *fp = fopen(filename, "rb");
	if(fp == NULL)
		return;

	char *str, *p, *section;
	section = NULL;
	while((str = freadline(fp)) != NULL)
	{
		// strip comments
		p = strpbrk(str, "#;");
		if(p) *p = 0;

		p = strchr(str, '=');
		if(p != NULL)
		{
			int n;
			char *name, *value;
			n = (int)(p - str) + 1;
			name = new char[n];
			value = new char[strlen(str) - n];

			*p++ = 0;
			for(; *p == ' ' || *p == '\t'; p++);
			strcpy(value, p);
			for(p = value + strlen(value) - 1;
				*p == ' ' || *p == '\t'; p--); p++;
			*p = 0;
			for(p = str; *p == ' ' || *p == '\t'; p++);
			strcpy(name, p);
			for(p = name + strlen(name) - 1;
				*p == ' ' || *p == '\t'; p--); p++;
			*p = 0;

			write_string(section, name, value);
			delete[] name;
			delete[] value;
		}
		else if((p = strchr(str, '[')) != NULL)
		{
			char *q = strchr(p, ']');
			if(q)
			{
				*q = 0;
				if(section) delete[] section;
				section = new char[(int)(q - p)];
				strcpy(section, p + 1);
			}
		}
		delete[] str;
	}
	if(section)
		delete[] section;
	fclose(fp);
}

void IniFile::write_string(const char *section, const char *name, const char *str)
{
	int n, n0, n1, n2;
	if(section == NULL) section = "";
	n0 = strlen(section);
	n1 = strlen(name);
	n2 = strlen(str);

	INIFILE_FIELDINFO *iffi = new INIFILE_FIELDINFO;
	iffi->buffer = new char[n0 + n1 + n2 + 3];
	strcpy(iffi->buffer, section);
	iffi->name = iffi->buffer + n0 + 1; strcpy(iffi->name, name);
	iffi->value = iffi->name + n1 + 1; strcpy(iffi->value, str);
	iffi->next = NULL;

	if(pList == NULL)
	{
		pList = iffi;
		return;
	}

	INIFILE_FIELDINFO *pprev = pList;
	while(pprev->next != NULL) pprev = pprev->next;
	pprev->next = iffi;
	iffi->prev = pprev;
}

const char* IniFile::readString(const char *section, const char *name, const char *def) const
{
	INIFILE_FIELDINFO *p;
	for(p = pList; p != NULL; p = p->next)
	{
		if(strcmp(p->buffer, section) == 0
		&& strcmp(p->name, name) == 0)
		{
			//printf("%s.%s='%s'\n", section, name, p->value);
			return p->value;
		}
	}
	//printf("%s.%s='%s'\n", section, name, def);
	return def;
}

bool IniFile::readBool(const char *section, const char *name, bool def) const
{
	const char *str = readString(section, name, NULL);
	if(str == NULL) return def;
	return atoi(str);
}

double IniFile::readDouble(const char *section, const char *name, double def) const
{
	const char *str = readString(section, name, NULL);
	if(str == NULL) return def;
	return atof(str);
}

int IniFile::readInteger(const char *section, const char *name, int def) const
{
	const char *str = readString(section, name, NULL);
	if(str == NULL) return def;
	return atoi(str);
}

u32 IniFile::readU32(const char *section, const char *name, u32 def) const
{
	u32 u;
	const char *str = readString(section, name, NULL);
	if(str == NULL) return def;
	sscanf(str, "%u", &u);
	return u;
}

const char* IniFile::expand(const char *name) const
{
	const char *p;

	p = strchr(name, '.');
	if(p)
	{
		int n = (int)(p - name);
		char *section = new char[n + 1];
		memcpy(section, name, n);
		section[n] = 0;

		p = readString(section, p + 1);
		if(p)
		{
			delete[] section;
			return p;
		}
	}
	return getenv(name);
}

bool char_check(int a)
{
	if(a <= 32 || a == ',' || a == ';') return false;
	return true;
}

int IniFile::readExpandedString(char *buffer, const char *section, const char *name) const{
	const char *str = readString(section, name);
	if(str == NULL)
	{
		if(buffer) buffer[0] = 0;
		return 1;
	}

	const char *p;
	char *q = buffer;
	for(p = str; *p != 0;)
	{
		if(*p == '$')
		{
			p++;
			const char *nm_start = p;
			if(*p == '(') // $(VAR) format
			{
				nm_start++;
				for(p++; *p != 0; p++)
					if(*p == ')') break;
				if(*p)
				{
					int n = (int)(p - nm_start);
					char *nm = new char[n + 1];
					memcpy(nm, nm_start, n);
					nm[n] = 0;

					const char *estr = expand(nm);
					if(estr)
					{
						n = strlen(estr);
						if(buffer) strcpy(q, estr);
						q += n;
					}
					p++;
					delete[] nm;
				}
				else // not balanced '('
				{
					if(buffer) strcpy(q, "$(");
					q += 2;
					p++;
				}
			}
			else // $VAR
			{
				for(p++; char_check(*p); p++);
				if(nm_start + 1 != p)
				{
					int n = (int)(p - nm_start);
					char *nm = new char[n + 1];
					memcpy(nm, nm_start, n);
					nm[n] = 0;

					const char *estr = expand(nm);
					if(estr)
					{
						n = strlen(estr);
						if(buffer) strcpy(q, estr);
						q += n;
					}
					p++;
					delete[] nm;
				}
				else
				{
					if(buffer) *q = '$';
					q++;
				}
			}
		}
		else
		{
			if(buffer) *q = *p++;
			q++;
		}
	}
	if(buffer) *q = 0;
	return (int)(q - buffer) + 1;
}

double IniFile::evaluateValue(const char *section, const char *name) const
{
	printf("IniFile::evaluateValue is not currently implemented.\n");
	return 0;
}

void IniFile::writeString(const char *section, const char *name, const char *str)
{
	write_string(section, name, str);
	fChanged = true;
}

void IniFile::writeBool(const char *section, const char *name, bool f)
{
	writeInteger(section, name, f);
}

void IniFile::writeDouble(const char *section, const char *name, double d)
{
	char buf[256];
	sprintf(buf, "%.8lg", d);
	writeString(section, name, buf);
}

void IniFile::writeInteger(const char *section, const char *name, int n)
{
	char buf[256];
	sprintf(buf, "%d", n);
	writeString(section, name, buf);
}

void IniFile::writeU32(const char *section, const char *name, u32 u)
{
	char buf[256];
	sprintf(buf, "%u", u);
	writeString(section, name, buf);
}

IniFile::~IniFile()
{
	if(fChanged)
		printf("Inifile write-back is not currently supported.\n");
}
