//--------------------------------------------------------------------------
// mpicomm.cpp
// GAPPA Parallel Distributed Environment Genetic Algorithm Library
// DGA MPI-Communicator Object
//--------------------------------------------------------------------------
// Copyright (C) 2000 Takashi Kawasaki, Tomoyuki Hiroyasu and Mitsunori Miki,
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright  notice
//    and this list of conditions.
// 2. Redistributions in binary form must reproduce the above  copyright notice
//    and this list of conditions in the documentation and/or other materials
//    provided with the distribution.
// 3. All advertising materials mentioning features or use of this software
//    must display the following acknowledgement:
//
//    This product includes software developed by Takashi Kawasaki,
//    Tomoyuki Hiroyasu and Mitsunori Miki except the functions of sgenrand()
//    and genrand().
//
// 4. The names of Takashi Kawasaki, Tomoyuki Hiroyasu and Mitsunori Miki
//    should not be used to endorse or promote products derived from this
//    software without specific prior written permission.

#include <vector>
#include <algorithm>
#include <mpi.h>
#include "mtrand.h"
#include "mpicomm.h"
#include "util.h"

#if defined(_MT)
#include "muthread.h"
#define LOCK()   commMutex.lock();
#define UNLOCK() commMutex.unlock();

#endif
#define LOCK()    /**/
#define UNLOCK() /**/
using namespace std;

MPICommunicator::MPICommunicator(MigrationRouter &mr, int *pargc, char ***pargv, const char *inifile, bool fSync) : Communicator(mr)

{
	MPI_Init(pargc, pargv);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &procs);
	resetCount = sessionCount = 0;
	nEnvs.resize(procs);
	afShutdown.resize(procs);

	wantHelp(!Communicator::loadSettings(*pargc, *pargv, inifile));
	argc = *pargc;
	argv = *pargv;
	if(inifile)
		filename = buffer_fn, strcpy(filename, inifile);
	else
		filename = NULL;

	// set communication buffer size to procs.
	// it initially has capasity to preserve the connections to all envs.
	isd.reserve(procs);

	fRunning = false;
	fSynchronizing = fSync;
}

MPICommunicator::~MPICommunicator()
{
	MPI_Finalize();
}

bool MPICommunicator::associate(GeneEnvironment *sender)
{
	if(fRunning) return false;
	envs.push_back(sender);
	Communicator::associate(sender);
}

bool MPICommunicator::disassociate(GeneEnvironment *sender)
{
	if(fRunning) return false;
	int i = getLocalEnvironmentId(sender);
	envs.erase(&envs[i]);
	Communicator::disassociate(sender);
	return true;
}

int MPICommunicator::getLocalEnvironmentId(GeneEnvironment *sender) const
{
	vector<GeneEnvironment *>::const_iterator it;
	it = find(envs.begin(), envs.end(), sender);
	if(it == envs.end())
		return -1;
	return it - envs.begin();
}

int MPICommunicator::getEnvironmentId(GeneEnvironment *sender) const
{
	return getLocalEnvironmentId(sender) + nEnvIdOffset;
}

void MPICommunicator::reset()
{
	if(fRunning)
		shutdown();

	if(resetCount < envs.size())
		resetCount++;

	if(resetCount == envs.size())
	{
		// get number of whole environments over MPI_COMM_WORLD
		MPI_Allgather(
			&resetCount, 1, MPI_INT,
			nEnvs.begin(), 1, MPI_INT,
			MPI_COMM_WORLD);

		int i;
		for(i = nEnvIdOffset = 0; i < rank; i++)
			nEnvIdOffset += nEnvs[i];

		nEnvironments = nEnvIdOffset;
		for(; i < procs; i++)
			nEnvironments += nEnvs[i];

		for(i = 0; i < resetCount; i++)
		{
			::loadSettings(NULL, envs[i],
				argc, argv, filename, nEnvIdOffset + i, true);
			envs[i]->setNumEnvironments(nEnvironments);
			envs[i]->setEnvironmentId(nEnvIdOffset + i);
		}

		// initializing router
		if(rank == 0) i = time(NULL);
		MPI_Bcast(&i, 1, MPI_INT, 0, MPI_COMM_WORLD);
		getRouter().srand(i);
		getRouter().resize(nEnvironments);
		getRouter().nextPattern();

		// clear shutdown flag
		for(i = 0; i < procs; i++)
			afShutdown[i] = false;

		resetCount = 0;
	}
}

void MPICommunicator::shutdown()
{
	// if isd.size() is not 0, it means this is shutdown and
	// we must do cleanup here.
	if(1) //isd.size())
	{
		int i, j, k;
		afShutdown[rank] = true;

		// sending shutdown signal
		for(i = 0; i < procs; i++)
			if(i != rank)
				send_async(new u8[1], 1, MPI_BYTE, i, 0);

		// check number of processors want to shutdown
		for(i = k = 0; i < procs; i++)
			k += afShutdown[i];

		// waiting for shutdown signal from other processors
		while(k < procs)
		{
			for(i = 0; i < procs; i++) if(!afShutdown[i])
			{
				for(j = 0; j < nEnvs[i]; j++)
				{
					int fArrival;
					MPI_Status stat;
					if(MPI_Iprobe(i, j, MPI_COMM_WORLD, &fArrival, &stat)
						== MPI_SUCCESS)
					{
						if(fArrival)
						{
							if(j == 0 && stat.count == 1)
								afShutdown[i] = true, k++;
							u8 *buffer = new u8[stat.count];
							MPI_Recv(buffer, stat.count, MPI_BYTE, i, j,
								MPI_COMM_WORLD, &stat);
							delete[] buffer; // take them away...
						}
					}
				}
			}
		}
		flush_buffer();
	}
	fRunning = false;

	// Wait for all processors
	MPI_Barrier(MPI_COMM_WORLD);

}

bool MPICommunicator::migration(GeneEnvironment *sender, const constgene_pointer apEmigrants[], int nEmigrants)
{
	fRunning = true;

	MigrationRouter& router = getRouter();
	int id = getEnvironmentId(sender);

	// if this is the first session in a new generation
	LOCK();
	if(sessionCount == 0)
		router.nextPattern();
	sessionCount = (sessionCount + 1) % envs.size();
	UNLOCK();

	int i, j, k;

	if(nEmigrants)
	{
		// get absolute destination id
		int dest = router[id];
		int dest_rank, dest_tag;
		for(dest_rank = i = 0;; dest_rank++)
		{
			if(dest < i + nEnvs[dest_rank])
			{
				dest_tag = dest - i;
				break;
			}
			i += nEnvs[dest_rank];
		}

		if(dest_rank == rank)
		{
			//--------------------------
			// Intra-node communication
			//--------------------------

		#if defined(_MT)
			envs[dest_tag].lockVectorGeneInstance();
		#endif

			for(i = 0; i < nEmigrants; i++)
				envs[dest_tag]->getVectorGeneInstance() = *apEmigrants[i];

		#if defined(_MT)
			envs[dest_tag].releaseVectorGeneInstance();
		#endif
		}
		else
		{
			//--------------------------
			// Inter-node communication
			//--------------------------

			int buffer_size = nEmigrants * sender->getRnaLength();
			u8 *buffer = new u8[buffer_size];
			for(i = 0; i < nEmigrants; i++)
				apEmigrants[i]->transcription(buffer + sender->getRnaLength() * i);

			LOCK();
			// data sending procedure
			send_async(buffer, buffer_size, MPI_BYTE, dest_rank, dest_tag);
			UNLOCK();
		}
	}

	int nJobs = 0;
	ASYNC_COMM_DATA **acd;
	acd = new ASYNC_COMM_DATA*[procs];

	//
	// Start Asynchronus Reading...
	//
	for(i = 0; i < procs; i++) if(!afShutdown[i] && i != rank)
	{
		acd[i] = new ASYNC_COMM_DATA[nEnvs[i]];
		for(j = 0; j < nEnvs[i]; j++)
		{
			// [i, j] means packet from Rank i to me (tag j)
			int fArrival, ret;
			MPI_Status stat;
			ret = MPI_Iprobe(i, j, MPI_COMM_WORLD, &fArrival, &stat);
			if(fArrival && ret == MPI_SUCCESS)
			{
				if(stat.count == 1)
				{
					u8 temp;
					MPI_Recv(&temp, 1, MPI_BYTE, i, j,
						MPI_COMM_WORLD, &stat);
					afShutdown[i] = true;
					afShutdown[rank] = true;
					acd[i][j].buffer = NULL;
				}
				else
				{
					acd[i][j].buffer = new u8[stat.count];
					MPI_Irecv(
						(void *)acd[i][j].buffer,
						stat.count, MPI_BYTE,
						i, j, MPI_COMM_WORLD, &acd[i][j].req);
					nJobs++;
				}
			}
			else
				acd[i][j].buffer = NULL;
		}
	}
	else
		acd[i] = NULL;

	//
	// Start Asynchronus Reading...
	//
	while(nJobs)
	{
		for(i = 0; i < procs; i++) if(i != rank)
		{
			for(j = 0; j < nEnvs[i]; j++) if(acd[i][j].buffer)
			{
				MPI_Status stat;
				int flag;
				MPI_Test(&acd[i][j].req, &flag, &stat);

				if(flag && stat.count)
				{
					int n = stat.count / sender->getRnaLength();
					for(k = 0; k < n; k++)
					{
						const u8 *data = acd[i][j].buffer + k * sender->getRnaLength();
					#if defined(_MT)
						envs[j]->lockVectorGeneInstance();
					#endif

						Gene& gen = envs[j]->getVectorGeneInstance();
						gen.reverseTranscription(data);

					#if defined(_MT)
						envs[j]->releaseVectorGeneInstance();
					#endif
					}
					delete[] acd[i][j].buffer; // take them away...
					acd[i][j].buffer = NULL;
					nJobs--;
				}
			}
		}
	}

	for(i = 0; i < procs; i++)
		if(acd[i]) delete[] acd[i];
	delete[] acd;

	check_buffer();

	// do synchronization
	if(fSynchronizing)
		MPI_Barrier(MPI_COMM_WORLD);

	// Must I shutdown now ?
	return afShutdown[rank];
}

void MPICommunicator::send_async(const void *buffer, int count, int type, int dest, int tag)
{
	ASYNC_COMM_DATA temp;
	temp.buffer = (const u8 *)buffer;
	MPI_Isend(
		const_cast<void *>(buffer),
		count, type, dest, tag, MPI_COMM_WORLD, &temp.req);
	isd.push_back(temp);
}

void MPICommunicator::check_buffer()
{
	vector<ASYNC_COMM_DATA>::iterator it;
	for(it = isd.end() - 1; it >= isd.begin(); it--)
	{
		MPI_Status stat;
		int flag;
		MPI_Test(&it->req, &flag, &stat);
		if(flag)
		{
			delete[] it->buffer;
			isd.erase(it);
		}
	}
}

void MPICommunicator::flush_buffer()
{
	MPI_Status stat;
	vector<ASYNC_COMM_DATA>::iterator it;
	for(it = isd.begin(); it < isd.end(); it++)
	{
		MPI_Wait(&it->req, &stat);
		delete[] it->buffer;
	}
	isd.erase(isd.begin(), isd.end());
}
