//--------------------------------------------------------------------------
// muthread.cpp
// GAPPA Parallel Distributed Environment Genetic Algorithm Library
// Thread Class Implementation for Win32/Pthreads
//--------------------------------------------------------------------------
// Copyright (C) 2000 Takashi Kawasaki, Tomoyuki Hiroyasu and Mitsunori Miki,
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright  notice
//    and this list of conditions.
// 2. Redistributions in binary form must reproduce the above  copyright notice
//    and this list of conditions in the documentation and/or other materials
//    provided with the distribution.
// 3. All advertising materials mentioning features or use of this software
//    must display the following acknowledgement:
//
//    This product includes software developed by Takashi Kawasaki,
//    Tomoyuki Hiroyasu and Mitsunori Miki except the functions of sgenrand()
//    and genrand().
//
// 4. The names of Takashi Kawasaki, Tomoyuki Hiroyasu and Mitsunori Miki
//    should not be used to endorse or promote products derived from this
//    software without specific prior written permission.

#include <iostream>
using namespace std;

#include "muthread.h"

// Entry Point for muThread::main, used internally
#if defined(USE_PTHREADS) // Pthreads
void * muThread::entry_point(_MU_MUTHREAD_CONTEXT *ptc)
{
	void *ret = ptc->pth->main(ptc->param);
#else // Windows
unsigned __stdcall muThread::entry_point(void *pv)
{
	_MU_MUTHREAD_CONTEXT *ptc = (_MU_MUTHREAD_CONTEXT *)pv;
	unsigned ret = (unsigned)ptc->pth->main(ptc->param);
#endif
	delete ptc;
	return ret;
}

// Constructor, do nothing now...
muThread::muThread()
{
	fRunning = false;
	th = 0;
}

// Execute thread, supposed to be called by others
bool muThread::execute(void *param)
{
	if(fRunning) return true;
	_MU_MUTHREAD_CONTEXT *ptc = new _MU_MUTHREAD_CONTEXT;
	ptc->pth = this;
	ptc->param = param;

#if defined(USE_PTHREADS) // Pthreads
	int ret = pthread_create(&th, NULL, entry_point, (void *)ptc);
	if(ret == 0)
		fRunning = true;
#else // Windows
	unsigned id;
	th = (void *)_beginthreadex(NULL, 0, entry_point, (void *)ptc, 0, &id);
	if(th)
		fRunning = true;
#endif
	if(!fRunning)
		delete ptc;
	return fRunning;
}

void muThread::exitThread(void *val)
{
	if(!isMyContext())
		return;

#if defined(USE_PTHREADS) // Pthreads
	pthread_exit(val);
#else // Windows
	_endthreadex((unsigned)val);
#endif
}

// current context is this thread ?
bool muThread::isMyContext()
{
	if(th == GET_MUTHREAD_CONTEXT())
		return true;
	return false;
}

// Thread Join Method, supposed to be called by others
void *muThread::join()
{
	if(!fRunning) return NULL; // if not running, do nothing

	if(isMyContext()) return NULL; // why call by myself ? no sence.

#if defined(USE_PTHREADS) // Pthreads
	void *pv;
	int ret = pthread_join(th, &pv);
	if(ret == 0)
		fRunning = false;
	return pv;
#else // Windows
	if(WaitForSingleObject(th, INFINITE) == WAIT_OBJECT_0)
	{
		fRunning = false;
		DWORD dw;
		GetExitCodeThread(th, &dw);
		CloseHandle(th);
		return (void *)dw;
	}
	return NULL;
#endif
}

// Destructor, simply call join for the purpose of gracefully shutdown.
muThread::~muThread()
{
	join();
}

//
// Ring Buffer Class
//

// Ring Buffer Constructor
muBuffer::muBuffer(int sz)
{
	buffer = new char[sz];
	w = r = 0;
	size = sz;
}

// Ring Buffer Destructor
muBuffer::~muBuffer()
{
	delete[] buffer;
}

int muBuffer::getQueueSize()
{
	mx.lock();
	int n = (w - r + size) % size;
	mx.unlock();
	return n;
}

void muBuffer::write(const void *data, int length)
{
	const char *p = (const char *)data;
	int l;

	while(length > 0)
	{
		mx.lock();
		if(w >= r)
		{
			l = size - w;
			if(length <= l)
			{
				memcpy(buffer + w, p, length);
				w += length;
				mx.unlock();
				return;
			}

			memcpy(buffer + w, p, l);
			length -= l;
			p += l;
			// w = 0; do nothing

			if(length <= r)
			{
				memcpy(buffer, p, length);
				w = length;
				mx.unlock();
				return;
			}

			memcpy(buffer, p, r);
			length -= r;
			p += r;
			w = r;
		}
		else
		{
			l = r - w;
			if(length <= l)
			{
				memcpy(buffer + w, p, length);
				w += length;
				mx.unlock();
				return;
			}
			memcpy(buffer + w, p, l);
			length -= l;
			p += l;
			w = r;
		}
		mx.unlock();
	}
}

void muBuffer::read(void *data, int length)
{
	char *p = (char *)data;
	int l;

	while(length > 0)
	{
		mx.lock();
		if(w >= r)
		{
			l = w - r;
			if(length <= l)
			{
				memcpy(p, buffer + r, length);
				r += length;
				mx.unlock();
				return;
			}
			memcpy(p, buffer + r, l);
			length -= l;
			p += l;
			r = w;
		}
		else
		{
			l = size - r;
			if(length <= l)
			{
				memcpy(p, buffer + r, length);
				r += length;
				mx.unlock();
				return;
			}
			memcpy(p, buffer + r, l);
			length -= l;
			p += l;
			r = 0;

			if(length <= w)
			{
				memcpy(p, buffer, length);
				r += length;
				mx.unlock();
				return;
			}
			memcpy(p, buffer, w);
			length -= w;
			p += w;
			r = w;
		}
		mx.unlock();
	}
}

int muBuffer::peek(void *data, int length)
{
	char *p = (char *)data;
	int l;

	mx.lock();
	if(w >= r)
	{
		l = w - r;
		if(length <= l)
		{
			memcpy(p, buffer + r, length);
			mx.unlock();
			return length;
		}
		memcpy(p, buffer + r, l);
		mx.unlock();
		return l;
	}

	l = size - r;
	if(length <= l)
	{
		memcpy(p, buffer + r, length);
		mx.unlock();
		return length;
	}
	memcpy(p, buffer + r, l);
	length -= l;
	p += length;

	if(length <= w)
	{
		memcpy(p, buffer, length);
		mx.unlock();
		return (int)(p - (char *)data);
	}
	memcpy(p, buffer, w);
	mx.unlock();
	return l + w;
}

#if defined(USE_WIN32_THREADS)
typedef unsigned (__stdcall *win32_entry_func)(void *);
#endif

void *ForkThreads::entry_point(THREAD_CX *ptc)
{
	void *pv = ptc->func(*ptc->pft, ptc->rank, ptc->param);
	ptc->rank = -1;
#if defined(USE_PTHREADS) // Pthreads
	pthread_exit(pv);
#else // Windows
	_endthreadex((unsigned)pv);
#endif
	return pv;
}

ForkThreads::ForkThreads(int n, fork_threads_entry_func func, void *param)
	: fJoined(false), sync(0), sz(n - 1)
{
	if(sz)
	{
		atc = new THREAD_CX[sz];

		int i;
		for(i = 0; i < sz; i++)
		{
			atc[i].rank = i + 1;
			atc[i].param = param;
			atc[i].func = func;
			atc[i].pft = this;
		}

		for(i = 0; i < sz; i++)
		{
#if defined(USE_PTHREADS) // Pthreads
			int ret = pthread_create(&atc[i].th, NULL, entry_point, (void *)&atc[i]);
			if(ret) atc[i].th = NULL;
#else // Windows
			unsigned id;
			atc[i].th
				= (void *)_beginthreadex(NULL, 0,
					(win32_entry_func)entry_point, (void *)&atc[i], 0, &id);
#endif
		}
	}

	ret0 = func(*this, 0, param);
}

void ForkThreads::join()
{
#if defined(USE_PTHREADS) // Pthreads
	int i;
	for(i = 0; i < sz; i++)
		if(atc[i].rank != -1)
			pthread_join(atc[i].th, &atc[i].ret);

#else // Windows
	int i, j;
	MUTHREAD_CONTEXT_T *aHandles = new MUTHREAD_CONTEXT_T[sz];
	for(i = j = 0; i < sz; i++)
		if(atc[i].rank != -1)
			aHandles[j++] = atc[i].th;
	if(j)
		MsgWaitForMultipleObjects(j, aHandles, TRUE, INFINITE, 0);

	for(i = 0; i < sz; i++)
	{
		GetExitCodeThread(atc[i].th, (DWORD *)&atc[i].ret);
		CloseHandle(atc[i].th);
	}
#endif
	fJoined = true;
}

ForkThreads::~ForkThreads()
{
	if(!fJoined) join();
	if(sz) delete[] atc;
}

void ForkThreads::synchronize() const
{
	lock();
	sync++; if(sync == sz) sync = 0;
	unlock();
	while(sync);
}
