//--------------------------------------------------------------------------
// muthread.h
// GAPPA Parallel Distributed Environment Genetic Algorithm Library
// Thread Class Implementation for Win32/Pthreads
//--------------------------------------------------------------------------
// Copyright (C) 2000 Takashi Kawasaki, Tomoyuki Hiroyasu and Mitsunori Miki,
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright  notice
//    and this list of conditions.
// 2. Redistributions in binary form must reproduce the above  copyright notice
//    and this list of conditions in the documentation and/or other materials
//    provided with the distribution.
// 3. All advertising materials mentioning features or use of this software
//    must display the following acknowledgement:
//
//    This product includes software developed by Takashi Kawasaki,
//    Tomoyuki Hiroyasu and Mitsunori Miki except the functions of sgenrand()
//    and genrand().
//
// 4. The names of Takashi Kawasaki, Tomoyuki Hiroyasu and Mitsunori Miki
//    should not be used to endorse or promote products derived from this
//    software without specific prior written permission.

#if !defined(_MU_MUTHREAD_XPLATFORM_H_)
#define _MU_MUTHREAD_XPLATFORM_H_

#if defined(_WIN32) && !defined(__CYGWIN__)
#define USE_WIN32_THREADS 1
#else
#define USE_PTHREADS 1
#endif

#if defined(USE_WIN32_THREADS)
// Windows
#if !defined(_MT)
#define _MT 1
#endif

#include <windows.h>
#include <process.h>
#else
// Pthreads
#include <pthread.h>
#endif

class muThread;

typedef struct {
	muThread *pth;
	void *param;
} _MU_MUTHREAD_CONTEXT;

//
// Mutex Implementation
//
#if defined(USE_PTHREADS) // Pthreads
typedef pthread_t MUTHREAD_CONTEXT_T;
typedef pthread_mutex_t MUTHREAD_MUTEX_T;
#define GET_MUTHREAD_CONTEXT() pthread_self()
#define MUTHREAD_MUTEX_CREATE(pmx) pthread_mutex_init((pmx), NULL)
#define MUTHREAD_MUTEX_DESTROY(pmx) pthread_mutex_destroy((pmx))
#define MUTHREAD_MUTEX_LOCK(pmx) pthread_mutex_lock((pmx))
#define MUTHREAD_MUTEX_UNLOCK(pmx) pthread_mutex_unlock((pmx))
#else // Windows
typedef HANDLE MUTHREAD_CONTEXT_T;
typedef CRITICAL_SECTION MUTHREAD_MUTEX_T;
#define GET_MUTHREAD_CONTEXT() GetCurrentThread()
#define MUTHREAD_MUTEX_CREATE(pmx) InitializeCriticalSection((pmx))
#define MUTHREAD_MUTEX_DESTROY(pmx) DeleteCriticalSection((pmx))
#define MUTHREAD_MUTEX_LOCK(pmx) EnterCriticalSection((pmx))
#define MUTHREAD_MUTEX_UNLOCK(pmx) LeaveCriticalSection((pmx))
#endif

class muMutex
{
public:
	muMutex() : owner(0) {MUTHREAD_MUTEX_CREATE(&mx);}
	~muMutex() {MUTHREAD_MUTEX_DESTROY(&mx);}
	void lock() {MUTHREAD_MUTEX_LOCK(&mx); owner = GET_MUTHREAD_CONTEXT();}
	void unlock() {owner = 0; MUTHREAD_MUTEX_UNLOCK(&mx);}
	inline bool isLocked() {return owner;}
	inline bool isMine() {return (GET_MUTHREAD_CONTEXT() == owner);}

private:
	MUTHREAD_CONTEXT_T owner;
	MUTHREAD_MUTEX_T mx;
};

//
// Thread Implementation
//
class muThread
{
public:
	muThread();

	bool isMyContext();
	bool execute(void *param);
	virtual ~muThread();
	void *join();
	bool isRunning() {return fRunning;}

protected:
	void exitThread(void *val);
	virtual void *main(void *arg) = 0;

private:
	MUTHREAD_CONTEXT_T th;
	bool fRunning;

#if defined(USE_PTHREADS) // Pthreads
	static void *entry_point(_MU_MUTHREAD_CONTEXT *tc);
#else // Windows
	static unsigned __stdcall entry_point(void *tc);
#endif
};

//
// Ring Buffer Implementation
//
class muBuffer
{
public:
	explicit muBuffer(int size = 30000);
	~muBuffer();
	int getQueueSize();
	void write(const void *data, int length);
	void read(void *buffer, int length);
	int peek(void *buffer, int length);

	// these method are difficult to deal with, but sometimes useful
	inline void lock() {mx.lock();}
	inline void unlock() {mx.unlock();}

private:
	muMutex mx;
	char *buffer;
	int size, r, w;
};

class ForkThreads;

typedef void * (*fork_threads_entry_func)(const ForkThreads& ft, int rank, void *param);

class ForkThreads
{
	int sz;
	void *ret0;
	bool fJoined;
	mutable volatile int sync;

	struct THREAD_CX
	{
		MUTHREAD_CONTEXT_T th;
		int rank;
		void *param;
		fork_threads_entry_func func;
		void *ret;
		const ForkThreads *pft;
	} *atc;

	static void *entry_point(THREAD_CX *ptc);

	mutable muMutex mutex;

public:
	ForkThreads(int n, fork_threads_entry_func func, void *param = NULL);
	void join();
	void *result(int n) {if(n) return atc[n - 1].ret; else ret0;}
	~ForkThreads();
	void lock() const {mutex.lock();}
	void unlock() const {mutex.unlock();}
	int size() const {return sz + 1;}
	void synchronize() const;
};

#endif /* _MU_MUTHREAD_XPLATFORM_H_ */
