//--------------------------------------------------------------------------
// util.cpp
// GAPPA Parallel Distributed Environment Genetic Algorithm Library
// utility module
//--------------------------------------------------------------------------
// Copyright (C) 2000 Takashi Kawasaki, Tomoyuki Hiroyasu and Mitsunori Miki,
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright  notice
//    and this list of conditions.
// 2. Redistributions in binary form must reproduce the above  copyright notice
//    and this list of conditions in the documentation and/or other materials
//    provided with the distribution.
// 3. All advertising materials mentioning features or use of this software
//    must display the following acknowledgement:
//
//    This product includes software developed by Takashi Kawasaki,
//    Tomoyuki Hiroyasu and Mitsunori Miki except the functions of sgenrand()
//    and genrand().
//
// 4. The names of Takashi Kawasaki, Tomoyuki Hiroyasu and Mitsunori Miki
//    should not be used to endorse or promote products derived from this
//    software without specific prior written permission.

#include <iostream>
#include <string>
#include <cstdio>
#include <cstdarg>

#include "util.h"

using namespace std;

bool loadLocalConfiguration(GeneEnvironment *env, IniFile& ini, const char *section)
{
	env->setMutationRate(
		ini.readDouble(section, "MutationRate", env->getMutationRate()));
	env->setCrossoverRate(
		ini.readDouble(section, "CrossoverRate", env->getCrossoverRate()));
	env->setNumCrossoverPoints(
		ini.readInteger(section, "CrossoverPoints", env->getNumCrossoverPoints()));
	env->setNumEliteSurvive(
		ini.readInteger(section, "Elite", env->getNumEliteSurvive()));
	env->setPopulationSize(
		ini.readInteger(section, "PopulationSize", env->getPopulationSize()));
	env->setEmigrationRate(
		ini.readDouble(section, "EmigrationRate", env->getEmigrationRate()));
	env->setNumEmigrants(
		ini.readInteger(section, "Emigrants", env->getNumEmigrants()));

	int mt = ini.readInteger(section, "MigrationTerm", -1);
	if(mt >= 0)
		env->setMigrationInterval(mt);
	else
		env->setMigrationInterval(
			ini.readInteger(section, "MigrationInterval",
				env->getMigrationTerm()));

	if(ini.readBool(section, "GenesisCompatible", false))
	{
		env->setScalingMethod(smByPseudoRank);
		env->setMutationMethod(mmGenesisCompatible);
	}

	const char *str = ini.readString(section, "MutationMethod", "");
	if(strcmp(str, "GenesisCompatible") == 0)
		env->setMutationMethod(mmGenesisCompatible);
	if(strcmp(str, "Normal") == 0)
		env->setMutationMethod(mmNormal);

	str = ini.readString(section, "ScalingMethod", "");
	if(strcmp(str, "ScalingWindow") == 0)
		env->setScalingMethod(smByScalingWin);
	if(strcmp(str, "Rank") == 0)
		env->setScalingMethod(smByRank);
	if(strcmp(str, "PseudoRank") == 0)
		env->setScalingMethod(smByPseudoRank);
	if(strcmp(str, "Normal") == 0)
		env->setScalingMethod(smNormal);

	env->enableEliteMigration(
		ini.readBool(section, "EliteMigration", env->enableEliteMigration()));

	return true;
}

bool loadConfiguration(GA_GLOBAL_SETTINGS *pggs, GeneEnvironment *env, const char *filename, const char *section, bool fSetGlobalSettings)
{
	IniFile ini(filename);
	return loadConfiguration(pggs, env, ini, section, fSetGlobalSettings);
}

bool loadGlobalSettings(GA_GLOBAL_SETTINGS *pggs, char *filename)
{
	return loadConfiguration(pggs, NULL, filename, 0, false);
}

bool loadConfiguration(GA_GLOBAL_SETTINGS *pggs, GeneEnvironment *env, IniFile& ini, const char *section, bool fSetGlobalSettings)
{
	GA_GLOBAL_SETTINGS ggs;

	if(pggs == NULL)
		pggs = &ggs;

	pggs->nGenerations = ini.readInteger("Global", "Generations", 1000);
	pggs->nEnvironments = ini.readInteger("Global", "Environments", 1);

	pggs->dvd.nBit = ini.readInteger("Global", "bits/var", 10);
	int clen = ini.readInteger("Global", "ChromoLength", 0);

	pggs->dvd.openMax = ini.readDouble("Global", "varOpenMax", 5.12);
	pggs->dvd.closedMin = ini.readDouble("Global", "varClosedMin", -5.12);

	pggs->nVariables = ini.readInteger("Global", "Variables", 1);
	if(clen > 0) // binary code GA
	{
		pggs->dvd.nBit = clen;
		pggs->nVariables = 1;
	}

	pggs->fCheckConvergence = ini.readBool("Global", "checkConvergence", 0);

	strcpy(pggs->function_2_load, ini.readString("Global", "Function", ""));
	pggs->fUseGrayCode = ini.readBool("Global", "useGrayCode", 0);
	pggs->nTrials = ini.readInteger("Global", "Trials", 1);

	strcpy(pggs->displayFormat,
		ini.readString("Global", "DisplayFormat", "%lf"));

	if(env)
	{
		loadLocalConfiguration(env, ini, "Global");
		if(section)
			loadLocalConfiguration(env, ini, section);

		if(fSetGlobalSettings)
		{
			env->setNumEnvironments(pggs->nEnvironments);
			env->setChromoLengthByBit(pggs->dvd.nBit * pggs->nVariables);
		}
	}
	return true;
}

bool loadConfiguration(GA_GLOBAL_SETTINGS *pggs, GeneEnvironment *env, IniFile& ini, int n, bool fSetGlobalSettings)
{
	char buf[256];
	sprintf(buf, "Local%d", n);
	return loadConfiguration(pggs, env, ini, buf, fSetGlobalSettings);
}

bool loadConfiguration(GA_GLOBAL_SETTINGS *pggs, GeneEnvironment *env, const char *filename, int n, bool fSetGlobalSettings)
{
	char buf[256];
	sprintf(buf, "Local%d", n);
	return loadConfiguration(pggs, env, filename, buf, fSetGlobalSettings);
}

bool loadCommandlineSetttings(GA_GLOBAL_SETTINGS *pggs, GeneEnvironment *env, int argc, const char * const argv[], bool fSetGlobalSettings)
{
	GA_GLOBAL_SETTINGS ggs;
	if(pggs == NULL) pggs = &ggs;
	GeneEnvironment temp;
	if(env == NULL) env = &temp;

	int i, rm, clen;
	double mut_rate = -1;
	clen = 0;
	for(i = 1; i < argc; i++)
	{
		if(*argv[i] == '-')
		{
			const char * const *param = &argv[i + 1];
			const char *_param = argv[i] + 2; // suboption
			rm = argc - i - 1;               // options remain
			switch(argv[i][1])
			{
			case 'G':
				if(rm)
				i++;
				break;

			case 'n':
				if(rm)
				{
					int n = atoi(param[0]);
					i++;
					switch(_param[0])
					{
					case 'P': // Population Size
						if(env) env->setPopulationSize(n);
						break;

					case 'G': // Generations
						pggs->nGenerations = n;
						break;

					case 'V': // number of variables
						pggs->nVariables = n;
						clen = pggs->nVariables * pggs->dvd.nBit;
						break;

					case 'T': // number of trials
						pggs->nTrials = n;
						break;

					case 'L': // chromosome length
						clen = n;
						break;

					case 'C': // number of crossover points
						if(env) env->setNumCrossoverPoints(n);
						break;

					case 'E':
						switch(_param[1])
						{
						case 'l': // number of elite
							if(env) env->setNumEliteSurvive(n);
							break;

						case 'm': // number of emigrants
							if(env) env->setNumEmigrants(n);
							break;

						case 0: // number of environments
							pggs->nEnvironments = n;
							break;
						}
						break;

					case 'M':
						switch(_param[1])
						{
						case 'T': // migration term
							if(env) env->setMigrationTerm(n);
						break;
						}
					}
				}
				break;

			case 'm':
				switch(_param[0])
				{
				case 'C': // check convergence
					pggs->fCheckConvergence = true;

				case 'N': // native mode
					if(env)
					{
						env->setScalingMethod(smNormal);
						env->setMutationMethod(mmNormal);
					}
					break;

				case 's': // scaling method
					{
						ScalingMethod sm = smNormal;
						switch(_param[1])
						{
						case 'S': sm = smByScalingWin; break;
						case 'R': sm = smByRank; break;
						case 'P': sm = smByPseudoRank; break;
						}
						if(env) env->setScalingMethod(sm);
					}
					break;

				case 'm': // mutation method
					{
						MutationMethod mm;
						if(_param[1] == 'G') mm = mmGenesisCompatible;
						else mm = mmNormal;
						if(env) env->setMutationMethod(mm);
					}
					break;

				case 'G':
					if(env)
					{
						env->setScalingMethod(smByPseudoRank);
						env->setMutationMethod(mmGenesisCompatible);
					}
					break;

				case 'E':
					if(env) env->enableEliteMigration(_param[2] != '-');
					break;
				}
				break;

			case 'r':
				if(rm)
				{
					double r = atof(param[0]);
					i++;
					switch(_param[0])
					{
					case 'm': // mutation rate
						mut_rate = r;
						break;

					case 'c': // crossover rate
						if(env) env->setCrossoverRate(r);
						break;

					case 'e': // emigration rate
						if(env) env->setEmigrationRate(r);
						break;
					}
				}
				break;

			case 'f':
				strcpy(pggs->displayFormat, param[0]);
				break;

			case 'F':
				strcpy(pggs->function_2_load, param[0]);
				break;

			case 'B':
				sscanf(param[0], "%d,%lf,%lf",
					&pggs->dvd.nBit, &pggs->dvd.closedMin, &pggs->dvd.openMax);
				clen = pggs->nVariables * pggs->dvd.nBit;
				break;

			case 'g':
				pggs->fUseGrayCode = true;
				break;

			case 'b':
				pggs->fUseGrayCode = false;
				break;

			case '-':
				if(strcmp(_param, "help"))
					break;

			case 'h':
			case 'H':
			case '?':
				return false;
			}
		}
	}

	if(env)
	{
		if(clen)
			env->setChromoLengthByBit(clen);
		if(mut_rate >= 0)
			env->setMutationRate(mut_rate);
	}
	return true;
}

bool loadSettings(GA_GLOBAL_SETTINGS *pggs, GeneEnvironment *env, int argc, const char * const argv[], const char *inifile, int n, bool fSetGlobalSettings)
{
	GA_GLOBAL_SETTINGS ggs;
	if(pggs == NULL) pggs = &ggs;
	if(inifile)
		loadConfiguration(pggs, env, inifile, n, fSetGlobalSettings);
	return loadCommandlineSetttings(pggs, env, argc, argv, fSetGlobalSettings);
}

static void make_fieldname(char *buf, const char *field)
{
	if(strlen(field) <= 2)
	{
		buf[0] = '-';
		buf[1] = field[0];
		buf[2] = field[1];
		buf[3] = 0;
	}
	else
	{
		buf[0] = buf[1] = '-';
		strcpy(buf + 2, field);
	}
}

const char * clParam(int argc, const char * const argv[], const char *field)
{
	int i;
	char buf[256];
	make_fieldname(buf, field);

	for(i = 1; i < argc; i++)
	{
		int rm = argc - i - 1;
		if(strcmp(argv[i], buf) == 0 && rm)
			return argv[i + 1];
	}
	return NULL;
}

bool clParamExists(int argc, const char * const argv[], const char *field)
{
	int i, l;
	char buf[256];
	make_fieldname(buf, field);

	for(i = 1; i < argc; i++)
		if(strcmp(argv[i], buf) == 0)
			return true;
	return false;
}

void mpi_trace(const char *format, ...)
{
	va_list argptr;
	char buffer[5000];
	int cnt;

	va_start(argptr, format);
	vsprintf(buffer, format, argptr);
	va_end(argptr);

	cerr << buffer;
}

const char *gappa_help =
"GAPPA Parallel Distributed Genetic Algorithm Library\n"
"Copyright (C) 2000 Takashi Kawasaki, Tomoyuki Hiroyasu and Mitsunori Miki,\n"
"All rights reserved.\n\n"
"[Command-line options]\n"
"-B <bit,min,max>  set characteristics of the all design variables\n"
"-F <function>     set function name (the behavior depends on implementation)\n"
"-f <format>       set display style by the 'printf' syntax\n"
"-b                using binary coding\n"
"-g                using gray coding\n"
"-mmG              using GENESIS compatible mutation\n"
"-mmN              using normal mutation\n"
"-msN              using simple scaling method\n"
"-msS              using Scaling-Window method\n"
"-msR              Scaling By Rank\n"
"-msP              Scaling by Pseudo-Rank(much faster than -msR)\n"
"-mN               using GAPPA native routines\n"
"-mG               using GENESIS compatible routines\n"
"-mE               using elite migration routines\n"
"-nG <num>         set number of generations to compute\n"
"-nP <num>         set sub population size (population per env.)\n"
"-nV <num>         set number of design variables\n"
"-nT <num>         set number of trials (depends on implementation)\n"
"-nL <num>         set chromosome length by value (do not use with -nV, -nB)\n"
"-nC <num>         set number of crossovering points (0: uniform-crossover)\n"
"-nE <num>         set number of environments(islands)\n"
"-nEl <num>        set number of elite\n"
"-nEm <num>        set number of emigrants(do not use with -re)\n"
"-nMT <num>        set migration interval\n"
"-rm <num>         set mutation rate (0.0 - 1.0)\n"
"-rc <num>         set crossover rate (0.0 - 1.0)\n"
"-re <num>         set emigration rate (0.0 - 0.5)\n";
